from thompcoutils.log_utils import get_logger
from thompcoutils.config_utils import ConfigManager
from thompcoutils.test_utils import assert_test
import logging
import sqlobject
import os


class DBConfig:
    def __init__(self, cfg_file, section):
        if section is None:
            section = 'DB CONNECTION'
        logger = get_logger()
        logger.debug('Initialing DbConfig')
        cfg_mgr = ConfigManager(cfg_file)
        self.ssh_tunnel = None
        self.type = cfg_mgr.read_entry(section, "type", default_value="mysql")
        self.sqlite_file = cfg_mgr.read_entry(section, "sqlite file", default_value="licenseserver.sqlite")
        self.db_host = cfg_mgr.read_entry(section, "db_host", default_value="localhost")
        self.db_username = cfg_mgr.read_entry(section, "db_username", default_value="my db user name")
        self.db_password = cfg_mgr.read_entry(section, "db_password", default_value="my db user password")
        self.schema = cfg_mgr.read_entry(section, "schema", default_value="my db schema")
        self.db_port = cfg_mgr.read_entry(section, "db_port", default_value=3306)
        self.rebuild = cfg_mgr.read_entry(section, "rebuild", default_value=False)
        self.ssh_username = cfg_mgr.read_entry(section, "ssh_username", default_value="")
        self.ssh_password = cfg_mgr.read_entry(section, "ssh_password", default_value="")
        self.ssh_host = cfg_mgr.read_entry(section, "ssh_host", default_value="localhost")
        self.use_ssh = cfg_mgr.read_entry(section, "use_ssh", default_value=False)
        self.ssh_timeout = cfg_mgr.read_entry(section, "ssh_timeout", default_value=5.0)
        self.ssh_tunnel_timeout = cfg_mgr.read_entry(section, "ssh_tunnel_timeout", default_value=5.0)
        if self.ssh_username == '':
            self.ssh_username = None
        if self.ssh_password == '':
            self.ssh_password = None


class DbUtils:
    def __init__(self, db_cfg_file=None, db_type=None, db_host=None, db_username=None, db_password=None, schema=None,
                 db_port=None, sqlite_file=None, ssh_username=None, ssh_password=None,
                 use_ssh=False, ssh_host=None, ssh_timeout=None, ssh_tunnel_timeout=None, section=None):
        logger = get_logger()
        self.connection = None
        self.ssh_tunnel = None
        self.tables = []
        self.db_type = db_type
        self.sqlite_file = sqlite_file
        self.db_host = db_host
        self.db_username = db_username
        self.db_password = db_password
        self.schema = schema
        self.db_port = db_port
        self.use_ssh = use_ssh
        self.ssh_host = ssh_host
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.ssh_timeout = ssh_timeout
        self.ssh_tunnel_timeout = ssh_tunnel_timeout

        if db_cfg_file is not None:
            db_cfg = DBConfig(cfg_file=db_cfg_file, section=section)
            self.db_type = db_cfg.type
            self.sqlite_file = db_cfg.sqlite_file
            self.db_host = db_cfg.db_host
            self.db_username = db_cfg.db_username
            self.db_password = db_cfg.db_password
            self.schema = db_cfg.schema
            self.db_port = db_cfg.db_port
            self.use_ssh = db_cfg.use_ssh
            self.ssh_username = db_cfg.ssh_username
            self.ssh_password = db_cfg.ssh_password
            self.ssh_timeout = db_cfg.ssh_timeout
            self.ssh_tunnel_timeout = db_cfg.ssh_tunnel_timeout
        if self.db_type == "sqlite":
            if self.sqlite_file is None:
                raise RuntimeError('"sqlite" requires "sqlite_file"')
            else:
                logger.debug("Using {} Sqlite database".format(self.sqlite_file))
                self._connect_sqlite(self.sqlite_file)
        elif self.db_type == "postgres":
            logger.debug('Using postgress database {} at {}'.format(self.schema, self.db_host))
            self._connect_postgres()
        elif "mysql" in self.db_type:
            logger.debug('Using mysql database {} at {}'.format(self.schema, self.db_host))
            self._connect_mysql()
        elif self.db_type == "odbc":
            raise RuntimeError("ODBC not implemented yet")
        else:
            raise RuntimeError("No database type selected")
        try:
            self.create_table(_TestTable)
            self.drop_tables([_TestTable])
        except Exception as e:
            logger.exception('Could not connect to database:{}'.format(e))

    def _connect_uri(self, uri):
        self.connection = sqlobject.sqlhub.processConnection = sqlobject.connectionForURI(uri)

    def _connect_sqlite(self, file_path):
        uri = "sqlite:" + file_path
        self._connect_uri(uri)

    def _connect_postgres(self):
        port_str = ""
        if self.db_port is not None:
            port_str = ":" + str(self.db_port)
        uri = "postgres://" + self.db_username + ":" + self.db_password + "@" + self.db_host + port_str + "/" + \
              self.schema
        self._connect_uri(uri)

    def _connect_mysql(self):
        import sshtunnel
        import mysql.connector
        logger = get_logger()
        if self.use_ssh:
            sshtunnel.SSH_TIMEOUT = self.ssh_timeout
            sshtunnel.TUNNEL_TIMEOUT = self.ssh_tunnel_timeout
            try:
                self.ssh_tunnel = sshtunnel.SSHTunnelForwarder(self.db_host,
                                                               ssh_username=self.ssh_username,
                                                               ssh_password=self.ssh_password,
                                                               remote_bind_address=(self.db_host, self.db_port))
                self.ssh_tunnel.daemon_forward_servers = True
                self.ssh_tunnel.start()
                self.connection = mysql.connector.connect(user=self.db_username, password=self.db_password,
                                                          host=self.ssh_host, port=self.ssh_tunnel.local_bind_port,
                                                          database=self.schema)
            except sshtunnel.BaseSSHTunnelForwarderError as e:
                logger.exception('Could not create an SSH Tunnel: {}'.format(str(e.value)))
                raise Exception(e)
            except mysql.connector.InterfaceError as e:
                logger.exception('Could not connect: {}'.format(e.msg))
                raise Exception(e)

        port_str = ""
        if self.db_port is not None:
            port_str = ":" + str(self.db_port)
        uri = "mysql://" + self.db_username + ":" + self.db_password + "@" + self.db_host + port_str + "/" + \
              self.schema
        self._connect_uri(uri)

    def add_table(self, table):
        self.tables.append(table)

    def set_tables(self, tables):
        self.tables = tables

    def shutdown(self):
        if self.ssh_tunnel is not None:
            self.ssh_tunnel.stop_flashing()

    def table_exists(self, table_name):
        return self.connection.tableExists(table_name.sqlmeta.table)

    def create_table(self, table):
        logger = get_logger()
        if not self.table_exists(table):
            logger.debug("creating table {}.{}".format(self.schema, str(table.sqlmeta.table)))
            table.createTable()
            logger.debug("created table {}.{}".format(self.schema, str(table.sqlmeta.table)))

    def create_tables(self, tables=None):
        logger = get_logger()
        logger.debug("Creating tables...")
        if tables is None:
            tables = self.tables
        for table in tables:
            self.create_table(table)

    def drop_tables(self, tables=None):
        logger = get_logger()
        logger.debug("Dropping tables...")
        if tables is None:
            tables = self.tables
        for table in tables:
            logger.debug("dropping table {}.{}".format(self.schema, str(table.sqlmeta.table)))
            table.dropTable(cascade=True, ifExists=True)
            logger.debug("dropped table {}".format(str(table.sqlmeta.table)))


class _TestTable(sqlobject.SQLObject):
    # noinspection PyPep8Naming,PyClassHasNoInit
    class sqlmeta:
        table = "Test_table"
    first_name = sqlobject.StringCol()
    last_name = sqlobject.StringCol()
    age = sqlobject.IntCol
    salary = sqlobject.FloatCol()

    @staticmethod
    def validate():
        _TestTable(first_name='Jordan', last_name='Thompson', age=35, salary=20.5)
        query = _TestTable.select(_TestTable.q.first_name == 'Jordan')
        assert_test(query.count() == 1, message='Ensuring count is 1', error='Count is not 1', passed='Count is 1')
        assert_test(query.getOne().last_name == 'Thompson', message='Ensuring data is inserted',
                    passed='Data inserted correctly')


class _TestDatabase(DbUtils):
    def __init__(self, db_cfg_file):
        super().__init__(db_cfg_file)
        self.set_tables([_TestTable])


def _main():
    config_file_name = "testing DbConfig only.ini"
    if not os.path.exists(config_file_name):
        cfg_mgr = ConfigManager(config_file_name, create=True)
        cfg_mgr.write(out_file=config_file_name)
        print('This should not print out!')
    test_db = _TestDatabase(db_cfg_file=config_file_name)
    if test_db.table_exists(_TestTable):
        test_db.drop_tables()
    assert_test(not test_db.table_exists(_TestTable),
                message='Testing that table {}.{} was successfully removed'.format(test_db.schema,
                                                                                   _TestTable.sqlmeta.table),
                error='Table {}.{} should not exist'.format(test_db.schema, _TestTable.sqlmeta.table),
                passed='Table {}.{} removed successfully'.format(test_db.schema, _TestTable.sqlmeta.table))
    test_db.create_tables()
    assert_test(test_db.table_exists(_TestTable),
                message='Testing that table {}.{} was successfully created'.format(test_db.schema,
                                                                                   _TestTable.sqlmeta.table),
                error='Table {}.{} should exist'.format(test_db.schema, _TestTable.sqlmeta.table),
                passed='Table {}.{} created successfully'.format(test_db.schema, _TestTable.sqlmeta.table))
    _TestTable.validate()
    test_db.shutdown()


if __name__ == "__main__":
    logging.basicConfig()
    logging.root.setLevel(logging.DEBUG)
    logging.getLogger('paramiko.transport').setLevel(logging.WARN)
    logging.root.propagate = True
    _main()



import base64
from Crypto.Cipher import AES
from Crypto import Random
import argparse
import hashlib


class Security:
    def __init__(self, key):
        self.bs = 32
        self.key = hashlib.sha256(key.encode()).digest()

    def encode(self, raw):
        raw = self._pad(raw)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(raw))

    def decode(self, enc):
        enc = base64.b64decode(enc)
        iv = enc[:16]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[16:]))

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s)-1:])]


def main():
    parser = argparse.ArgumentParser(description="Encrypts/Decrypts a string",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("--key",
                        required=True,
                        help="AES key to use for encryption")
    parser.add_argument("--value",
                        required=True,
                        help="value to encrypt or decrypt")
    parser.add_argument("--dir",
                        default="encrypt",
                        choices=["encrypt", "decrypt"],
                        help="Direction of encryption")
    args = parser.parse_args()
    sec = Security(args.key)
    if args.dir == "encrypt":
        print(sec.encode(args.value))
    else:
        print(sec.decode(args.value))


if __name__ == '__main__':
    main()



seconds_per_minute = 60
minutes_per_hour = 60
hours_per_day = 24
milliseconds_per_second = 1000
seconds_per_hour = seconds_per_minute * minutes_per_hour
seconds_per_day = seconds_per_hour * hours_per_day
microseconds_per_second = 1000 * milliseconds_per_second



from thompcoutils.test_utils import assert_test
import datetime

_debug = False
_now_time = None


def _now():
    if _debug:
        return _now_time
    else:
        return datetime.datetime.now()


class CronTimeException(Exception):
    pass


class CronTime:

    def __init__(self, day_of_week=0, month=0, day_of_month=0, hour=0, minute=0):
        self.day_of_week = day_of_week
        self.month = month
        self.day_of_month = day_of_month
        self.hour = hour
        self.minute = minute

    def __str__(self):
        return CronTime.strfcrontime(self, '%w %d %m %H %M')

    # allowed characters are:
    # %w single digit weekday (0=Sunday)
    # %d two digit day of the month
    # %m two digit month
    # %H two digit hour
    # %M two digit minute
    @staticmethod
    def strfpcrontime(source_string, format_string):
        cron_time = CronTime()
        j = 0
        i = 0
        for r in range(0, len(format_string)):
            if i > len(format_string) - 1:
                break
            if format_string[i] == '%':
                fmt = format_string[i+1]
                if fmt == 'w':
                    cron_time.day_of_week = int(source_string[j:j+1])
                    i += 1
                elif fmt == 'm':
                    cron_time.month = int(source_string[j:j+2])
                elif fmt == 'd':
                    cron_time.day_of_month = int(source_string[j:j+2])
                elif fmt == 'H':
                    cron_time.hour = int(source_string[j:j+2])
                elif fmt == 'M':
                    cron_time.minute = int(source_string[j:j+2])
                else:
                    raise CronTimeException("%{} not recognised as a CronTime format".format(fmt))
            i += 1
            j += 1
        return cron_time

    def strfcrontime(self, format_string):
        rtn = ''
        i = 0
        for y in range(0, len(format_string)):
            if i > len(format_string) - 1:
                break
            if format_string[i] == '%':
                fmt = format_string[i+1]
                if fmt == 'w':
                    rtn += '{:1d}'.format(self.day_of_week)
                elif fmt == 'm':
                    rtn += '{:02d}'.format(self.month)
                elif fmt == 'd':
                    rtn += '{:02d}'.format(self.day_of_month)
                elif fmt == 'H':
                    rtn += '{:02d}'.format(self.hour)
                elif fmt == 'M':
                    rtn += '{:02d}'.format(self.minute)
                else:
                    raise CronTimeException("%{} not recognised as a CronTime format".format(fmt))
                i += 1
            else:
                rtn += format_string[i]
            i += 1
        return rtn

    def triggered(self):
        minute = False
        hour = False
        day = False
        month = False
        day_of_week = False

        if self.minute == 0:
            minute = True
        elif _now().minute == self.minute:
            minute = True

        if self.hour == 0:
            hour = True
        elif _now().hour == self.hour:
            hour = True

        if self.day_of_month == 0:
            day = True
        elif _now().day == self.day_of_month:
            day = True

        if self.month == 0:
            month = True
        elif _now().month == self.month:
            month = True

        if self.day_of_week == 0:
            day_of_week = True
        elif _now().weekday() == self.day_of_week:
            day_of_week = True

        return minute and hour and day and month and day_of_week


def _test_crontime():
    format_string = '%w %d %m %H %M'
    source_string = '1 02 03 04 05'
    cron_time = CronTime.strfpcrontime(source_string, format_string)
    assert_test(cron_time.day_of_week == 1)
    assert_test(cron_time.day_of_month == 2)
    assert_test(cron_time.month == 3)
    assert_test(cron_time.hour == 4)
    assert_test(cron_time.minute == 5)
    print(cron_time)
    test_string = cron_time.strfcrontime(format_string)
    assert_test(test_string == source_string)


if __name__ == "__main__":
    _test_crontime()



import inspect
import logging
import logging.config
import logging.handlers
import os


class LogUtilsMissingConfigFileException(Exception):
    pass


class LogUtilsMissingFileOrFolderException(Exception):
    pass

class LogUtilsConfigFileFormatException(Exception):
    pass

class RelativePathRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, relative_path, file_name, max_bytes=2000, backup_count=100):
        local_path = os.path.dirname(os.path.abspath(__file__))
        log_path = os.path.join(local_path, relative_path)
        if not os.path.isdir(log_path):
            os.mkdir(log_path)
        log_file_name = os.path.join(log_path, file_name)
        super(RelativePathRotatingFileHandler, self).__init__(log_file_name, max_bytes, backup_count)


def get_log_file_name():
    for handler in logging.root.handlers:
        if handler.baseFilename is not None:
            return handler.baseFilename
    return None


def get_logger(additional_name=None):
    stack = inspect.stack()
    if len(stack) == 2:
        file_name = os.path.basename(stack[1][1]).split(".")[0]
    else:
        file_name = os.path.basename(stack[2][1]).split(".")[0]
    the_function = stack[1][3]
    # noinspection PyBroadException
    try:
        if len(stack[1][0].f_locals) > 0:
            the_class = str(stack[1][0].f_locals["self"].__class__.__name__) + "."
        else:
            the_class = ""
    except Exception:
        the_class = ""
    if additional_name is None:
        additional_name = ''
    logger_name = "{}.{}{}{}".format(file_name, the_class, the_function, additional_name)
    return logging.getLogger(logger_name)


def _get_arg_list(kwargs):
    string = None
    count = 0
    if len(kwargs) == 1:
        # noinspection PyBroadException
        try:
            string = kwargs["msg"]
        except Exception:
            pass
    if string is None:
        string = ""
        for key in kwargs:
            val = kwargs[key]
            string += str(key) + "=" + str(val)
            if count < len(kwargs) - 1:
                string += ","
                count += 1
    return string


def current_function_name():
    stack = inspect.stack()
    the_function = stack[1][3]
    return the_function


def start_function(logger, **kwargs):
    stack = inspect.stack()
    the_function = stack[1][3]
    logger.debug("Starting {}({})".format(the_function, _get_arg_list(kwargs)))


def end_function(logger, **kwargs):
    stack = inspect.stack()
    the_function = stack[1][3]
    logger.debug("Ending {}({})".format(the_function, _get_arg_list(kwargs)))


class _Test:
    @staticmethod
    def function1():
        logger = get_logger()
        logger.debug("debug in function1")
        logger.info("info in function2")
        logger.warning("warning in function1")
        logger.error("error in function1")
        logger.critical("critical in function1")


def _test_function3():
    logger = get_logger()
    val1 = 5
    val2 = 6
    logger.debug("debug in _test_function3")
    logger.info("info in _test_function3")
    logger.warning("warning in _test_function3")
    logger.error("error in _test_function3")
    logger.critical("critical in _test_function3")
    start_function(logger, msg="_test_function3")
    start_function(logger, val1=val1, val2=val2)
    val1 = 10
    val2 = {"first": 1, "second": 2}
    end_function(logger, val1=val1, val2=val2)


def _test_function2():
    logger = get_logger()
    val1 = 5
    val2 = 6
    logger.debug("debug in _test_function2")
    logger.info("info in _test_function2")
    logger.warning("warning in _test_function2")
    logger.error("error in _test_function2")
    logger.critical("critical in _test_function2")
    start_function(logger, msg="_test_function2")
    start_function(logger, val1=val1, val2=val2)
    val1 = 10
    val2 = {"first": 1, "second": 2}
    end_function(logger, val1=val1, val2=val2)


def _test_function():
    logger = get_logger()
    val1 = 5
    val2 = 6
    logger.debug("debug in _test_function")
    logger.info("info in _test_function")
    logger.warning("warning in _test_function")
    logger.error("error in _test_function")
    logger.critical("critical in _test_function")
    start_function(logger, msg="_test_function")
    start_function(logger, val1=val1, val2=val2)
    val1 = 10
    val2 = {"first": 1, "second": 2}
    end_function(logger, val1=val1, val2=val2)


def load_log_config(config_file_name):
    logger = get_logger()
    if not os.path.exists(config_file_name):
        raise LogUtilsMissingConfigFileException("Log configuration file {} does not exist.  "
                                                 "Consider creating one for better debugging".format(config_file_name))
    try:
        logging.config.fileConfig(config_file_name)
    except KeyError as e:
        raise LogUtilsConfigFileFormatException('log file format exception:{}'.format(str(e)))
    logger.debug("Beginning logging with configuration from:{}".format(config_file_name))


def _main():
    logger = get_logger()
    start_function(logger, msg="starting")
    test = _Test()
    test.function1()
    _test_function()
    _test_function2()
    _test_function3()


if __name__ == '__main__':
    load_log_config('../tests/logging.ini')
    _main()



import datetime
import calendar
from dateutil.relativedelta import relativedelta, TH, MO


class Holiday(datetime.datetime):
    THANKSGIVING = "thanksgiving"
    CHRISTMAS = "christmas"
    NEW_YEARS = "newyears"
    INDEPENDENCE = "independence"
    MEMORIAL = "memorial"
    LABOR = "labor"

    def __new__(cls, holiday_in):
        year = datetime.datetime.now().year
        holiday = holiday_in.lower()

        if holiday == Holiday.THANKSGIVING:
            # fourth thursday in November
            month = 11
            november = datetime.datetime(year=year, month=month, day=1)
            day = (november + relativedelta(day=31, weekday=TH(-1))).day
        elif holiday == Holiday.CHRISTMAS:
            month = 12
            day = 25
        elif holiday == Holiday.NEW_YEARS or holiday == "new years":
            month = 1
            day = 1
        elif holiday == Holiday.INDEPENDENCE:
            month = 7
            day = 4
        elif holiday == Holiday.MEMORIAL:
            month = 5
            day = 31
        elif holiday == Holiday.LABOR:
            # first monday in September
            month = 9
            november = datetime.datetime(year=year, month=month, day=1)
            day = (november + relativedelta(day=1, weekday=MO(1))).day
        else:
            raise RuntimeError("{} is not a recognised Holiday".format(holiday))
        return super().__new__(cls, year=year, month=month, day=day)


def is_weekend(dt):
    return dt.weekday() >= 5



import datetime
from thompcoutils.log_utils import get_logger
import threading
import sys
import units as units
from queue import Queue
from enum import Enum
import time
from pio_exception import PioException

from threading_utils import WorkerLoopThread, WorkerThread

# noinspection PyBroadException
try:
    import RPi.GPIO as GPIO
    import Adafruit_ADS1x15
    import Adafruit_DHT
    import smbus
    GPIO.setwarnings(False)
    GPIO.setmode(GPIO.BCM)
    pio_loaded = True
    get_logger().debug('GPIO successfully loaded')
except Exception as e:
    pio_loaded = False
    raise PioException('GPIO not loaded')


class ADC:
    class SensorType(Enum):
        ADS1115 = 0
        ADS1015 = 1

    class Values:
        def __init__(self, values, voltages):
            """
            Values are retrieved by the _get_data_thread and loaded into a Values object that is then pushed onto the
            queue.
            :param values: array of values from the ADC
            :param voltages: array of voltages from the ADC
            """
            self.time_stamp = datetime.datetime.now()
            self.values = values
            self.voltages = voltages

    @staticmethod
    def _get_data_thread(adc):
        logger = get_logger()
        """
        _get_data_thread retrieves the data values from the adc and pushes them onto the thread.  It is called via a
        thread and runs in a loop until stopped.
        :param adc: the ADC (should be self)
        :return: None
        """
        values = [0] * 4
        voltages = [0] * 4
        for i in range(4):
            values[i] = adc.adc.read_adc(i, gain=adc.gain)
            v2 = values[i] * adc.multiplier
            v3 = ((adc.r1 * v2) / adc.r2) + v2
            voltages[i] = v3
            if v3 > adc.last_voltage[i] + adc.tolerance or v3 < adc.last_voltage[i] - adc.tolerance:
                append = True
            adc.last_voltage[i] = v3
        # print('| {0:>6} | {1:>6} | {2:>6} | {3:>6} |'.format(*values))
        logger.debug('| {0:.2f} | {1:.2f} | {2:.2f} | {3:.2f} |'.format(*voltages))
        adc.queue.put_nowait(adc.Values(values, voltages))
        time.sleep(adc.update)

    def __init__(self, chip_type=SensorType.ADS1115, gain=1, r1=1000000, r2=220000, tolerance=.1, update=.25):
        """
        An Analog to Digital Converter (ADC) converts analog values to digital ones.
        :param chip_type: the type of the chip (from ADC.SensorType)
        :param gain: gain required to amplify the signal
        :param r1: R1 from a resistor voltage splitter
        :param r2: R2 from a resistor voltage splitter
        :param tolerance: amount of tolerance allowed
        :param update: amount of time to sleep between  adc readings
        """
        self.gains = {1: 4.096, 2: 6.144, 3: 6.144, 4: 1.024, 8: 0.512, 16: 0.256}
        self.gain = gain
        self.r1 = r1
        self.r2 = r2
        self.tolerance = tolerance
        self.update = update
        self.queue = Queue(1000)
        self.last_voltage = [0] * 4
        self.worker_thread = None

        # Create an ADS1115 ADC (16-bit) instance.
        if chip_type == ADC.SensorType.ADS1115:
            self.adc = Adafruit_ADS1x15.ADS1115()
            self.resolution_range = {min: -32768, max: 32767}
        elif chip_type == ADC.SensorType.ADS1015:
            self.adc = Adafruit_ADS1x15.ADS1015(address=0x49, busnum=1)
            self.resolution_range = {min: -2048, max: 2047}
        else:
            raise RuntimeError("Invalid Chip Type:{}".format(chip_type))
        self.multiplier = 2 * self.gains[gain] / (self.resolution_range[max] - self.resolution_range[min])

    def start(self):
        """
        Starts the ADC._get_data_thread loop
        """
        if self.worker_thread and self.worker_thread.is_running():
            return

        # empty the queue
        while not self.queue.empty():
            self.queue.get_nowait()
        self.worker_thread = WorkerLoopThread(callback_function=ADC._get_data_thread, parameters=self)
        self.worker_thread.start()

    def stop(self):
        """
        Stops the ADC._get_data_thread loop
        """
        if self.worker_thread and self.worker_thread.is_running():
            self.worker_thread.stop()
            self.worker_thread = None

    # noinspection PyProtectedMember
    def is_running(self):
        """
        Indicates the ADC._get_data_thread is running
        :return: True or False if running/not running
        """
        return self.worker_thread and self.worker_thread.is_running()


class Button:
    def __init__(self, pin, callback, pull_up_down=GPIO.PUD_UP, package=None):
        """
        Represents a button on the pi.
        check out this article on debouncing:
        https://www.raspberrypi.org/forums/viewtopic.php?t=134394
        :param pin: pin the button is on
        :param callback: callback to call when pressed
        :param pull_up_down: defines if the button is pulled up or down with a resistor
        :param package: package to be passed to the callback function (generally self)
        """
        self.pin = pin
        self.package = package
        self.last_push = datetime.datetime.now()
        self.callback = callback
        GPIO.setup(pin, GPIO.IN, pull_up_down=pull_up_down)
        GPIO.add_event_detect(pin, edge=GPIO.BOTH, callback=self._debounce_function)

    def _debounce_function(self, pin):
        """
        This function debounces the button.  Buttons are inherently noisy (they ring when pressed)  This function waits
        a period of time before declaring the button to have settled into the current state.
        :param pin: pin the button is on
        :return: None
        """
        time_now = datetime.datetime.now()
        current_state = GPIO.input(self.pin)
        if (time_now - self.last_push).microseconds > .1 * units.microseconds_per_second:
            self.callback(pin, current_state, self.package)
        self.last_push = time_now


class TimedButton:
    def __init__(self, pin, callback, pull_up_down=GPIO.PUD_UP, package=None):
        """
        A TimedButton initiates a callback after a button has been pressed and then released.  It passes the duration
        the button was pressed to the callback
        :param pin: pin the button is on
        :param callback: callback to call when pressed
        :param pull_up_down: indicates the button is pulled up or down with a resistor
        :param package: package to pass to the callback
        """
        self.pin = pin
        self.package = package
        self.last_push = datetime.datetime.now()
        self.callback = callback
        self.press_time = None
        GPIO.setup(pin, GPIO.IN, pull_up_down=pull_up_down)
        GPIO.add_event_detect(pin, edge=GPIO.BOTH, callback=self._debounce_function)

    def _debounce_function(self, pin):
        """
        This function debounces the button.  Buttons are inherently noisy (they ring when pressed)  This function waits
        a period of time before declaring the button to have settled into the current state.
        :param pin: pin the button is on
        :return: None
        """
        time_now = datetime.datetime.now()
        current_state = GPIO.input(self.pin)
        if (time_now - self.last_push).microseconds > .1 * units.microseconds_per_second:
            if current_state and self.press_time is not None:
                self.callback(pin, datetime.datetime.now() - self.press_time, self.package)
            else:
                self.press_time = datetime.datetime.now()
        self.last_push = time_now


class PiOut:
    def __init__(self, pin, initial_state=None):
        """
        Defines an  output pin for the pi
        :param pin: Pin to control
        :param initial_state: True (up) or False (down)
        """
        if pio_loaded:
            GPIO.setup(pin, GPIO.OUT)
        self.pin = pin
        self.state = False
        if initial_state is not None:
            self.toggle(initial_state)

    def toggle(self, new_state=None):
        """
        Toggles the current state of new_state is not provided
        :param new_state: state to force the pin to
        :return: None
        """
        logger = get_logger()
        if new_state is None:
            self.state = not self.state
        else:
            self.state = new_state
        logger.debug("pin {} toggled {}".format(self.pin, self.state))
        if pio_loaded:
            GPIO.output(self.pin, self.state)

    def turn_on(self):
        """
        Turns on the output to this PiOut's pin
        :return: None
        """
        GPIO.output(self.pin, True)

    def turn_off(self):
        """
        Turns off the output to this PiOut's pin
        :return: None
        """
        GPIO.output(self.pin, False)

    def is_on(self):
        """
        Indicates whether the pin is high or low
        :return: True if high, False if low
        """
        return self.state


class Led(PiOut):
    def __init__(self, pin, initial_state=None):
        """
        Basic LED.
        :param pin: Pin the LED is on
        :param initial_state: True for on, False for off
        """
        PiOut.__init__(self, pin, initial_state)


class Relay(PiOut):
    def __init__(self, pin, initial_state=None):
        """
        Relay is a very basic PIO device
        :param pin: Pin the relay is on
        :param initial_state: True for on, False for off
        """
        PiOut.__init__(self, pin, initial_state)


class FlashingLed(Led):
    def __init__(self, pin, on_off, count=None, initial_state=None):
        """
    Flashing LED is an LED that flashes
        :param pin: Pin the relay is on
        :param on_off: time (milliseconds) the led is [on, off]
        :param count: Number of times to flash (None is forever)
        :param initial_state:
        """
        PiOut.__init__(self, pin, initial_state)
        self.on_off = on_off
        self.count = count
        self.worker_thread = None

    # noinspection PyProtectedMember
    def _flasher(self):
        """
        This is the callback that flashes the LED on and off.  It is called once by start()
        :return: None
        """
        logger = get_logger()
        self.toggle(True)
        logger.debug("FlashingLight pin {} is {}".format(self.pin, self.state))
        time.sleep(self.on_off[0])
        if self.worker_thread._is_running:
            self.toggle(False)
        logger.debug("FlashingLight pin {} is {}".format(self.pin, self.state))
        time.sleep(self.on_off[1])

    def start(self):
        """
        This method starts the LED flashing.
        :return: None
        """
        if self.worker_thread is None:
            self.worker_thread = WorkerThread(callback_function=self._flasher)
            self.worker_thread.start()

    def stop(self):
        """
        This method stops the LED from flashing
        :return: None
        """
        if self.worker_thread is not None:
            self.worker_thread.stop_flashing()
            self.worker_thread = None


class ColorLed:
    def __init__(self, red_pin, green_pin, blue_pin):
        """
        ColorLed is a color led that can be flashed.
        :param red_pin: pin number for red
        :param green_pin: pin number for green
        :param blue_pin: pin number for blue
        """
        self._red = Led(red_pin)
        self._green = Led(green_pin)
        self._blue = Led(blue_pin)
        self.turn_off()
        self.on_off = []
        self.colors = []
        self._flasher_running = False
        self._flashing = False

    def set_colors(self, colors):
        """
        Sets the colors for the led
        :param colors: array of boolean [red, green, blue]
        :return: None
        """
        self.colors = colors

    def turn_on(self):
        """
        Turns on the colors with the colors set in set_colors()
        :return: None
        """
        self._red.toggle(self.colors[0])
        self._green.toggle(self.colors[1])
        self._blue.toggle(self.colors[2])

    def red(self):
        """
        Turns the led red
        :return: None
        """
        self.set_colors([False, True, True])
        self.turn_on()

    def green(self):
        """
        Turns the led green
        :return: None
        """
        self.set_colors([True, False, True])
        self.turn_on()

    def blue(self):
        """
        turns the led blue
        :return: None
        """
        self.set_colors([True, True, False])
        self.turn_on()

    def yellow(self):
        """
        turns the led yellow using red & green
        :return: None
        """
        self.set_colors([False, False, True])
        self.turn_on()

    def magenta(self):
        """
        turns the led magenta using red & blue
        :return: None
        """
        self.set_colors([False, True, False])
        self.turn_on()

    def cyan(self):
        """
        turns the led cyan using blue & green
        :return: None
        """
        self.set_colors([True, False, False])
        self.turn_on()

    def white(self):
        """
        turns the led white using red, green, & blue
        :return: None
        """
        self.set_colors([False, False, False])
        self.turn_on()

    def turn_off(self):
        """
        turns the led off
        :return: None
        """
        self.set_colors([True, True, True])
        self.turn_on()

    # noinspection PyProtectedMember
    @staticmethod
    def _flasher(color_led):
        """
        this callback changes the led to be off if it is on, and on if it is off.  It is called once and runs in a loop.
        :param color_led: the led to be flashed
        :return: None
        """
        on = True
        original_colors = color_led.colors
        print("starting flasher loop with {} {} {}".format(color_led.colors,
                                                           color_led.on_off,
                                                           threading.get_ident()))
        if color_led._flasher_running:
            print('already flashing!')
            return
        color_led._flashing = True
        color_led._flasher_running = True
        while color_led._flashing:
            if on:
                color_led.set_colors(original_colors)
                color_led.turn_on()
                time.sleep(color_led.on_off[0]/1000)
            else:
                color_led.turn_off()
                time.sleep(color_led.on_off[1]/1000)
            on = not on
        color_led._flasher_running = False
        print("exiting flasher loop with {} {} {}".format(color_led.colors,
                                                          color_led.on_off,
                                                          threading.get_ident()))

    def start_flashing(self, on_off):
        """
        starts the color led flashing
        :param on_off: time (milliseconds) the led is [on, off]
        :return: None
        """
        self.on_off = on_off
        WorkerThread(callback_function=self._flasher, parameters=self).start()

    def stop_flashing(self, fast_return=False):
        """
        stops the flashing LED and waits until it is done flashing.
        NOTE that this will pause execution until the current flash tops or fast_return is True
        :return: None
        """
        self._flashing = False  # stop flashing
        while not fast_return and self._flasher_running:
            time.sleep(.01)


class ButtonLightSequence:
    @staticmethod
    def callback(pin, current_state, button_light_sequence):
        """
        Callback when the button is pressed down
        :param current_state: not used
        :param pin: not used
        :param button_light_sequence: the ButtonLightcountdown
        :return: None
        """
        print("button pressed down")
        button_light_sequence.stepping = True
        while button_light_sequence.stepping:
            button_light_sequence.state_light.state_stepper()
            print("sleeping for {}".format(button_light_sequence.time_between_states/1000))
            time.sleep(button_light_sequence.time_between_states/1000)

    def __init__(self, callback, button_pin, states, time_between_states=1000):
        """
        A ButtonLightCountdown starts a sequence when the button is pressed.  As the button remains pressed down, the
        squence continues.  When the  button is released, the sequence is stopped and the callback is called.
        :param callback: callback to call when the button is released
        :param button_pin: pin the button is on
        :param states: list of StateLight.States
        """
        self.callback = callback
        self.button = Button(pin=button_pin, callback=ButtonLightSequence.callback, package=self)
        self.state_light = StateLight(states)
        self.time_between_states = time_between_states
        self.stepping = False


class StateLight:
    class State:
        def __init__(self, light, duration=0, colors=None, on_off=None, package=None):
            """
            A StateLight.State is a state of an individual light.  It can be an Led, ColoredLed, or a FlashingLed.
            :param light: the light for this state
            :param duration: the duration it will be on this state.  Don't use this parameter if you are going to call
            state_stepper manually
            :param on_off: the time the led will be [on, off] in milliseconds
            :param colors: the array of colors if this is a ColoredLed
            """
            if colors is None:
                colors = []
            self.light = light
            self.duration = duration
            self.colors = colors
            self.on_off = on_off
            self.package = package

    def __init__(self, state_list):
        """
        Initializes a StateLight which is one or more LEDs defined by StateLight.States
        :param state_list: array of all of the StateLight.States
        """
        self.state_list = state_list
        self.stepper_worker_thread = None
        self.state = 0

    def state_stepper(self):
        """
        Steps to the next state of the light(s)
        :return: None
        """
        # activate the state
        state = self.state_list[self.state]
        # test if goes in loop
        if isinstance(state.light, FlashingLed):
            state.light.start()
        elif isinstance(state.light, Led):
            state.light.toggle(True)
        elif isinstance(state.light, ColorLed):
            state.light.stop_flashing()
            state.light.set_colors(state.colors)
            state.light.start_flashing(on_off=state.on_off)
        time.sleep(state.duration)
        # advance the counter
        if self.state >= len(self.state_list) - 1:
            self.state = 0
        else:
            self.state += 1
        # turn off the state
        if isinstance(state.light, FlashingLed):
            state.light.stop()
            while state.light.state:
                time.sleep(.01)
        elif isinstance(state.light, Led):
            state.light.toggle(new_state=False)

    def start(self):
        """
        Starts the StateLight so that it will automatically go through all of the states
        :return: None
        """
        if self.stepper_worker_thread is None:
            self.stepper_worker_thread = WorkerThread(self.state_stepper)
            self.stepper_worker_thread.start()

    def stop(self):
        """
        Stops the StateLight
        :return:
        """
        if self.stepper_worker_thread is not None:
            self.stepper_worker_thread.stop_flashing()
            self.stepper_worker_thread = None
        for state in self.state_list:
            if isinstance(state.light, FlashingLed):
                state.light.stop()
            elif isinstance(state.light, Led):
                state.light.toggle(new_state=False)

    def get_state(self):
        """
        Gets the current StateLight.State and if it is running
        :return: current StateLight.State, True/False
        """
        return self.state, self.is_running

    # noinspection PyProtectedMember
    def is_running(self):
        """
        Gets if the thread for this StateLight is running
        :return: True/False if the Statelight is running
        :return: True/False if the Statelight is running
        """
        if self.stepper_worker_thread is None:
            return False
        else:
            return self.stepper_worker_thread._is_running


class TemperatureHumidity:
    class SensorType(Enum):
        """
        Defines the sensor type
        """
        AM2302 = 0

    class Values:
        def __init__(self, temperature, humidity):
            """
            Values recorded at the specified time
            :param temperature: Temperature recorded
            :param humidity: Humidity recorded
            """
            self.time_stamp = datetime.datetime.now()
            self.temperature = temperature
            self.humidity = humidity
            self.worker_thread = None

    @staticmethod
    def _get_data_thread(temp_humidity):
        """
        This function is  continuously called to get data from the sensor.  It fills the queue with Values
        :param temp_humidity: this is the calling object (self)
        :return: None
        """
        logger = get_logger()
        humidity, temperature = Adafruit_DHT.read_retry(temp_humidity.sensor_type, temp_humidity.pin)
        if temp_humidity.queue.full():
            logger.critical('Temperature and Humidity queue is full')
            return
        if humidity is not None and temperature is not None:
            if not temp_humidity.centigrade:
                temperature = 9.0 / 5.0 * temperature + 32
            logger.debug("pin:{}, temp:{}, humidity:{}".format(temp_humidity.pin, temperature, humidity))
            temp_humidity.queue.put(TemperatureHumidity.Values(temperature, humidity))

    def __init__(self, pin, sensor_type, centigrade=False):
        """
        Initializes a TemperatureHumidity sensor
        :param pin: Pin the sensor is on
        :param sensor_type: Type of sensor (see TemperatureHumidity.SensorType)
        :param centigrade: Set to True if you want centigrade values
        """
        self.pin = pin
        self.queue = Queue(1000)
        self.centigrade = centigrade
        if sensor_type == TemperatureHumidity.SensorType.AM2302:
            self.sensor_type = Adafruit_DHT.AM2302
        else:
            raise RuntimeError('Invalid Sensor Type')
        self.worker_thread = None

    def start(self):
        """
        This method starts the thread to read data from the sensor.  Note that it may take some time to stabilize
        """
        if self.worker_thread and self.worker_thread.is_running():
            return

        # empty the queue
        while not self.queue.empty():
            self.queue.get_nowait()
        self.worker_thread = WorkerLoopThread(callback_function=self._get_data_thread, parameters=self)
        self.worker_thread.start()

    def stop(self):
        """
        This method stops the thread.
        """
        if self.worker_thread and self.worker_thread.is_running():
            self.worker_thread.stop()
            self.worker_thread = None

    def is_running(self):
        """
        Indicates if the thread is running
        :return: True if the data collection thread is running
        """
        return self.worker_thread and self.worker_thread.is_running()


class VibrationSensor:
    class Runner(threading.Thread):
        def __init__(self, parent_sensor):
            super(VibrationSensor.Runner, self).__init__()
            self.parent = parent_sensor
            logger = get_logger()
            logger.debug("creating a VibrationSensor.Runner")
            self._is_running = False
            self.jitter = 1000  # 1 mS
            self.changed_time = None
            self.changed_state = False
            self.last_time = datetime.datetime.now()

        def run(self):
            logger = get_logger()
            logger.debug("starting VibrationSensor.Runner")
            self._is_running = True
            while self._is_running:
                if self.changed_time is not None:
                    time_delay = (self.changed_time - self.last_time).microseconds
                    self.last_time = self.changed_time
                    logger.critical("delay:{}".format(time_delay))
                    if time_delay > self.jitter:
                        logger.critical("moving")

        # noinspection PyUnusedLocal
        def state_change(self, pin):
            get_logger().critical("called")
            self.changed_state = datetime.datetime.now()

    def __init__(self, pin, callback):
        self.pin = pin
        self.callback_function = callback
        self.thread = None
        self.last_change = datetime.datetime.now()
        self.last_state = False
        self.runner = None

    def start(self):
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.pin, GPIO.IN)
        self.runner = VibrationSensor.Runner(self)
        self.runner.start()
        GPIO.add_event_detect(self.pin, GPIO.RISING, bouncetime=1)
        GPIO.add_event_callback(self.pin, self.runner.state_change)

    def _vibration_detected(self, pin):
        logger = get_logger()
        time_now = datetime.datetime.now()
        current_state = GPIO.input(pin)
        logger.critical("called:{}, {}, {}, {}".format(current_state, time_now, self.last_change,
                                             (time_now - self.last_change).microseconds))
        if current_state != self.last_state:
            if (time_now - self.last_change).microseconds > 1000:
                logger.critical("time exceeded")
                if current_state:
                    logger.critical("movement")
                else:
                    logger.critical("no movement")
                self.last_state = current_state
                self.last_change = time_now


class Accelerometer:
    # ADXL345 constants
    EARTH_GRAVITY_MS2 = 9.80665
    SCALE_MULTIPLIER = 0.004

    DATA_FORMAT = 0x31
    BW_RATE = 0x2C
    POWER_CTL = 0x2D

    BW_RATE_1600HZ = 0x0F
    BW_RATE_800HZ = 0x0E
    BW_RATE_400HZ = 0x0D
    BW_RATE_200HZ = 0x0C
    BW_RATE_100HZ = 0x0B
    BW_RATE_50HZ = 0x0A
    BW_RATE_25HZ = 0x09

    RANGE_2G = 0x00
    RANGE_4G = 0x01
    RANGE_8G = 0x02
    RANGE_16G = 0x03

    MEASURE = 0x08
    AXIS_DATA = 0x32

    address = None

    def __init__(self, address=0x53,
                 bandwidth_rate_flag=BW_RATE_100HZ, range_flag=RANGE_2G):
        self.address = address
        if pio_loaded:
            # select the correct i2c bus for this revision of Raspberry Pi
            revision = ([line[12:-1]
                         for line in open('/proc/cpuinfo', 'r').readlines()
                         if line[:8] == "Revision"] + ['0000'])[0]
            self.bus = smbus.SMBus(1 if int(revision, 16) >= 4 else 0)
        self.bus.write_byte_data(self.address, Accelerometer.BW_RATE, bandwidth_rate_flag)
        value = self.bus.read_byte_data(self.address, Accelerometer.DATA_FORMAT)
        value &= ~0x0F
        value |= range_flag
        value |= 0x08
        self.bus.write_byte_data(self.address, Accelerometer.DATA_FORMAT, value)
        self.bus.write_byte_data(self.address, Accelerometer.POWER_CTL, Accelerometer.MEASURE)

    def get_axis(self, g_force=False):
        data = self.bus.read_i2c_block_data(self.address, Accelerometer.AXIS_DATA, 6)
        now = datetime.datetime.now()
        x = data[0] | (data[1] << 8)
        if x & (1 << 16 - 1):
            x -= 1 << 16
        y = data[2] | (data[3] << 8)
        if y & (1 << 16 - 1):
            y -= 1 << 16
        z = data[4] | (data[5] << 8)
        if z & (1 << 16 - 1):
            z -= 1 << 16
        if g_force:
            x *= Accelerometer.SCALE_MULTIPLIER
            y *= Accelerometer.SCALE_MULTIPLIER
            z *= Accelerometer.SCALE_MULTIPLIER
        else:
            x *= Accelerometer.EARTH_GRAVITY_MS2
            y *= Accelerometer.EARTH_GRAVITY_MS2
            z *= Accelerometer.EARTH_GRAVITY_MS2
        x = round(x, 4)
        y = round(y, 4)
        z = round(z, 4)
        return {"x": x, "y": y, "z": z, "time": now}


class Motor:
    def __init__(self, accelerometer, resting_values):
        self.accelerometer = accelerometer
        self.resting_values = resting_values
        self.movement_checker = None
        self.motor_running = False

    class MovementChecker(threading.Thread):
        check_frequency = .001
        value_count = 10

        def __init__(self, parent_motor, callback):
            super(Motor.MovementChecker, self).__init__()
            self.callback = callback
            self.parent_motor = parent_motor
            self.thread_is_running = False
            self.values = [{"x": None, "y": None, "z": None}] * self.value_count

        def run(self):
            last_entry = 0
            filled = False
            resting_called = running_called = False
            while self.thread_is_running:
                # keep the counter within this thread's internal buffer
                if last_entry == len(self.values) - 1:
                    last_entry = 0
                    filled = True
                else:
                    last_entry += 1
                self.values[last_entry] = self.parent_motor.accelerometer.get_axis(True)
                if filled:  # no checking occurs until the small array is filled with movement data
                    if self.is_resting():
                        self.parent_motor.motor_running = False
                        if not resting_called:  # have not called stopped callback
                            running_called = False
                            resting_called = True
                            threading.Thread(target=self.callback, args=(False, self.when_event_occurred()))
                    else:  # motor is moving
                        self.parent_motor.motor_running = True
                        if not running_called:  # have not called callback
                            running_called = True
                            resting_called = False
                            threading.Thread(target=self.callback, args=(True, self.when_event_occurred()))
                time.sleep(Motor.MovementChecker.check_frequency)

        def is_resting(self):
            avg_x = avg_y = avg_z = 0
            for value in self.values:
                avg_x += value["x"]
                avg_y += value["y"]
                avg_z += value["z"]
            avg_x /= len(self.values)
            avg_y /= len(self.values)
            avg_z /= len(self.values)
            if avg_x > self.parent_motor.resting_values["x"] or \
               avg_y > self.parent_motor.resting_values["y"] or \
               avg_z > self.parent_motor.resting_values["z"]:
                return True
            else:
                return False

        def when_event_occurred(self):
            for value in self.values:
                if self.parent_motor.motor_running:  # look for when the motor first started moving
                    if value["x"] > self.parent_motor.resting_values["x"] or \
                       value["y"] > self.parent_motor.resting_values["y"] or \
                       value["z"] > self.parent_motor.resting_values["z"]:
                        return value["time"]
                else:  # look for when the motor first stopped moving
                    if value["x"] < self.parent_motor.resting_values["x"] or \
                       value["y"] < self.parent_motor.resting_values["y"] or \
                       value["z"] < self.parent_motor.resting_values["z"]:
                        return value["time"]

    def start(self, callback):
        Motor.MovementChecker(self, callback).start()

    @staticmethod
    def get_resting_values(accelerometer, duration_in_mils):
        start_time = datetime.datetime.now()
        min_x = min_y = min_z = sys.float_info.max
        max_x = max_y = max_z = 0
        while True:
            values = accelerometer.get_axis(True)
            if values["x"] > max_x:
                max_x = values["x"]
            if values["x"] < min_x:
                min_x = values["x"]
            if values["y"] > max_y:
                max_y = values["y"]
            if values["y"] < min_y:
                min_y = values["y"]
            if values["z"] > max_z:
                max_z = values["z"]
            if values["z"] < min_z:
                min_z = values["z"]
            if (datetime.datetime.now() - start_time).total_seconds() * 1000 >= duration_in_mils:
                break
            time.sleep(.001)

        return {"x": {"max": max_x, "min": min_x},
                "y": {"max": max_y, "min": min_y},
                "z": {"max": max_z, "min": min_z}
                }

    def is_moving(self):
        return self.motor_running



import sys
import os
import unittest
import logging
import datetime
from shutil import copyfile
from thompcoutils.cron_time import CronTime
from thompcoutils.test_utils import assert_test
from thompcoutils.config_utils import ConfigManager, EmailConnectionConfig, HiLow, PiConfig

test_path = 'test_ini_files'
if not os.path.exists(test_path):
    os.mkdir(test_path)
log_configuration_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.ini')
logging.config.fileConfig(log_configuration_file)


def test_replace(filename, old_string, new_string):
    # Safely read the input filename using 'with'
    try:
        with open(filename) as f:
            s = f.read()
            if old_string not in s:
                print('"{old_string}" not found in {filename}.'.format(**locals()))
                return

        # Safely write the changed content, if found in the file
        with open(filename, 'w') as f:
            print('Changing "{old_string}" to "{new_string}" in {filename}'.format(**locals()))
            s = s.replace(old_string, new_string)
            f.write(s)
    except Exception as e:
        raise e


def test_hi_low_vals(file_name, section, value):
    hi_low = HiLow(file_name=file_name, value_type=float)
    hi_value = value
    places = 10
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(values[HiLow.change_amount_tag] > 1, "value should be large")

    diff = 1
    value -= diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Down, "Should be moving Down")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.hi_tag] == hi_value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.NoChange, "Should not be moving")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.hi_tag] == hi_value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(values[HiLow.change_amount_tag] == 0, "value has not changed")

    diff = 2
    value += diff
    values = hi_low.write_value(entry=section, value=value)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    diff = 1.1
    value += diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    diff = 10.1
    value -= diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Down, "Should be moving Down")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    value = 14
    vals = hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(vals == values, "values should match")


class TestConfigUtils(unittest.TestCase):
    def test_config_mgr(self):
        file_name = os.path.join(test_path, "test_config.ini")
        for write in [True, False]:
            if write:
                if os.path.isfile(file_name):
                    os.remove(file_name)
            else:
                test_replace(file_name, "Rover", "Baily")
            cfg_mgr = ConfigManager(file_name,
                                    "This is the title of the ini file\n"
                                    "You can have multiple lines if you use line breaks", write)
            first = cfg_mgr.read_entry("User 1", "date_time", datetime.datetime.now())
            second = cfg_mgr.read_entry("User 1", "first name", "Joe", "This is the first name")
            last = cfg_mgr.read_entry("User 1", "last name", "Brown", "This is the last name")
            age = cfg_mgr.read_entry("User 1", "age", 12)
            is_male = cfg_mgr.read_entry("User 1", "male", True)
            weight = cfg_mgr.read_entry("User 1", "weight", 23.5)
            values = cfg_mgr.read_entry("User 1", "values", {"height": 7.5, "weight": 10, "name": "Fred"})
            weights = cfg_mgr.read_entry("User 1", "weights", [23.5, 22])
            names = cfg_mgr.read_entry("User 1", "names", ["Joe", "Fred"])
            cfg_mgr.write_entry("User 1", "male", False)
            cfg_mgr.write_entry("User 1", "parent", "Fred")
            cfg_mgr.write_entry("User 1", "date_time", datetime.datetime.now())
            cfg_mgr.write_entry("User 1", "cron_time", CronTime(day_of_month=1, hour=2, minute=3),
                                format_string='%d %H %M')
            # section = cfg_mgr.read_section("user 2", {"first name": "Sally",
            #                                           "last name": "Jones",
            #                                           "age": 15,
            #                                           "is_male": False,
            #                                           "weight": 41.3},
            #                                "You only get to add notes at the top of the section using this method")
            if write:
                test1 = cfg_mgr.read_entry("User 1", "dog name", "Rover")
                assert_test(test1 == "Rover", "value should be Rover")
            else:
                test1 = cfg_mgr.read_entry("User 1", "dog name", "Rover")
                assert_test(test1 == "Baily", "value should be Rover")
                test2 = cfg_mgr.read_entry("User 1", "cat name", "Tinkerbell", use_default_if_missing=False)
                assert_test(test2 is None, "missing value should be none")
                val = cfg_mgr.read_entry("User 1", "cron_time", CronTime(day_of_month=1, hour=2, minute=3))
                assert_test(val.day_of_month == 1)
                assert_test(val.day_of_week == 0)
                assert_test(val.month == 0)
                assert_test(val.hour == 2)
                assert_test(val.minute == 3)

            print(first)
            print(second)
            print(last)
            print(age)
            print(is_male)
            print(weight)
            print(values)
            print(weights)
            print(names)
            # print(section)
            if write:
                test_file = file_name
                cfg_mgr.write(out_file=test_file, overwrite=False, stop=False)
                contents = open(test_file, "r")
                print("File contents are:")
                print("====================================================")
                print(contents.read())
                print("====================================================")
                contents.close()

    def test_email_connection_config(self):
        file_name = os.path.join(test_path, "test_email.ini")
        for write in [True, False]:
            if write:
                if os.path.isfile(file_name):
                    os.remove(file_name)
            else:
                pass
            cfg_mgr = ConfigManager(file_name,
                                    "This is the title of the ini file\n"
                                    "You can have multiple lines if you use line breaks", write)
            email_connection = EmailConnectionConfig(cfg_mgr)
            print(email_connection.username)
            print(email_connection.password)
            print(email_connection.from_user)
            print(email_connection.smtp_host)
            print(email_connection.smtp_port)
            if write:
                test_file = file_name
                cfg_mgr.write(out_file=test_file, overwrite=True, stop=False)
                contents = open(test_file, "r")
                print("File contents are:")
                print("====================================================")
                print(contents.read())
                print("====================================================")
                contents.close()

    def test_hi_low(self):
        file_name = os.path.join(test_path, "test_hi_low.ini")
        if os.path.isfile(file_name):
            os.remove(file_name)
        value = 50
        test_hi_low_vals(file_name, "first", value)
        value = 60
        test_hi_low_vals(file_name, "second", value)
        value = 40
        test_hi_low_vals(file_name, "third", value)

    @staticmethod
    def remove_file(f):
        if os.path.exists(f):
            os.unlink(f)

    def test_pi_config(self):
        if sys.platform == 'linux' and os.uname()[4][:3] == 'arm':
            # prep the files
            config_filename = '/tmp/configuration.ini'
            default_config_filename = '{}{}'.format(config_filename, PiConfig._default_file_extension)
            # set the "/boot" folder and wpa_supplicant file for testing on any machine
            PiConfig._boot_folder = "/tmp"
            PiConfig._wpa_supplicant_file = '/tmp/wpa_supplicant.config'
            PiConfig._host_name_file = '/tmp/host'
            PiConfig._hosts_file = '/tmp/hosts'
            # First remove any files from previous testing
            self.remove_file(config_filename)
            self.remove_file(default_config_filename)
            self.remove_file(PiConfig._wpa_supplicant_file)
            self.remove_file(PiConfig._host_name_file)
            self.remove_file(PiConfig._hosts_file)
            copyfile('/etc/hosts', PiConfig._hosts_file)
            f = open(PiConfig._host_name_file, 'w')
            f.write('old_hostname\n')
            f.close()
            f = open(PiConfig._wpa_supplicant_file, 'w')
            f.write('ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev\n')
            f.write('update_config=1\n')
            f.write('country=US\n')
            f.write('network={\n')
            f.write('\tssid="Your network name/SSID"\n')
            f.write('\tpsk="Your WPA/WPA2 security key"\n')
            f.write('\tkey_mgmt=WPA-PSK\n')
            f.write('}\n')
            f.close()
            # test default values
            assert_test(value=not os.path.exists(default_config_filename), message='{} should not exist'.format(default_config_filename),
                        error='{} should not exist'.format(default_config_filename))
            assert_test(value=not os.path.exists(config_filename), message='{} should not exist'.format(config_filename),
                        error='{} should not exist'.format(config_filename))
            pi_config = PiConfig()
            email_config = EmailConnectionConfig(pi_config)
            for wifi in pi_config.wifi_sections:
                print(wifi.section)
                print(wifi.ssid)
                print(wifi.wifi_key)
                print(wifi.hidden)
            print(email_config.username)
            print(email_config.password)
            # save the default file
            pi_config.write()
            assert_test(value=os.path.exists(default_config_filename), message='testing creation of {}'.format(default_config_filename),
                        error='{} should exist'.format(default_config_filename))
            assert_test(value=not os.path.exists(config_filename), message='testing removal of {}'.format(config_filename),
                        error='{} should not exist'.format(config_filename))
            # fill the test config file
            new_host_name = 'new_host_name'
            new_country_code = 'CA'
            wifi_ssid = 'testing wifi'
            wifi_key = 'testing wifi key'
            f = open(config_filename, 'w')
            f.write('[misc]\n')
            f.write('host_name = {}\n'.format(new_host_name))
            f.write('wifi_country_code = {}\n\n'.format(new_country_code))
            for i in range(1, 3):
                f.write('[wifi_{}]\n'.format(i))
                f.write('ssid = {} {}\n'.format(wifi_ssid, i))
                f.write('key = {} {}\n'.format(wifi_key, i))
                f.write('hidden = {}\n'.format(True if i == 1 else False))
                f.write('priority = {}\n\n'.format(i))
            f.close()
            # read the test config file
            assert_test(value=os.path.exists(default_config_filename), message='{} should exist'.format(default_config_filename),
                        error='{} should exist'.format(default_config_filename))
            pi_config = PiConfig()
            assert_test(pi_config.misc.host_name==new_host_name)
            assert_test(pi_config.misc.wifi_country_code==new_country_code)
            assert_test(value=os.path.exists(default_config_filename), message='testing creation of {}'.format(default_config_filename),
                        error='{} should exist'.format(default_config_filename))
            assert_test(value=not os.path.exists(config_filename), message='testing removal of {}'.format(config_filename),
                        error='{} should not exist'.format(config_filename))
            assert_test(pi_config.misc.host_name == new_host_name)
            for i in range(0, len(pi_config.wifi_sections)):
                section = pi_config.wifi_sections[i]
                if '1' in section.section:
                    section_id = 1
                elif '2' in section.section:
                    section_id = 2
                else:
                    section_id = 3
                assert_test(section.ssid.startswith(wifi_ssid))
                assert_test(section.wifi_key.startswith(wifi_key))
                assert_test(section.hidden == (True if section_id == 1 else False))
                assert_test(section.priority == section_id)
            pi_config.write_wpa_supplicant()
            pi_config.write_host_files()
            # test no actual file but default file exists
            assert_test(value=os.path.exists(default_config_filename), message='{} should exist'.format(default_config_filename),
                        error='{} should exist'.format(default_config_filename))
            assert_test(value=not os.path.exists(config_filename), message='{} should not exist'.format(config_filename),
                        error='{} should not exist'.format(config_filename))
            PiConfig()
            assert_test(value=os.path.exists(default_config_filename), message='testing creation of {}'.format(default_config_filename),
                        error='{} should exist'.format(default_config_filename))
            assert_test(value=not os.path.exists(config_filename), message='testing removal of {}'.format(config_filename),
                        error='{} should not exist'.format(config_filename))
            print('Done')




from thompcoutils.threading_utils import ThreadManager, Watchdog, WorkerThread, WorkerLoopThread
from thompcoutils.log_utils import current_function_name
import time
import datetime
import unittest

_time_format = '%H:%M:%S'


def function_test1(name, value1, value2):
    print("thread {} starting at {}".format(name, datetime.datetime.now().strftime(_time_format)))
    time.sleep(2)
    print("thread {} done at {}".format(name, datetime.datetime.now().strftime(_time_format)))
    value1 *= 5
    value2 *= 2
    return value1, value2


def function_test2(name, value1):
    print("thread {} starting at {}".format(name, datetime.datetime.now().strftime(_time_format)))
    time.sleep(1)
    print("thread {} done at {}".format(name, datetime.datetime.now().strftime(_time_format)))
    value1 *= 3
    return value1


def _timeout_function(watchdog):
    print('{} Test timeout function called'.format(datetime.datetime.now()))
    watchdog.stop()


def worker_function():
    for i in range(0,5):
        print('{} running _worker_thread'.format(datetime.datetime.now()))
        time.sleep(1)
    print('WorkerThread finishing')


class TestThreadUtils(unittest.TestCase):
    def setUp(self):
        pass

    def atest_watchdog_1(self):
        print('{} {} starting'.format(datetime.datetime.now(), current_function_name()))
        watchdog = Watchdog(2, _timeout_function)
        watchdog.start()
        for i in range(0, 2):
            time.sleep(1)
            print('{} tickling'.format(datetime.datetime.now()))
            watchdog.tickle()
        time.sleep(2)
        print('{} watchdog_test done...'.format(datetime.datetime.now()))

    def atest_worker_thread(self):
        print('{} {} starting'.format(datetime.datetime.now(), current_function_name()))
        worker_thread = WorkerThread(worker_function)
        worker_thread.start()
        print('{} {} done'.format(datetime.datetime.now(), current_function_name()))

    def atest_threads(self):
        print('{} {} starting'.format(datetime.datetime.now(), current_function_name()))
        ThreadManager('one', function_test1, 'one', 1, 2)
        ThreadManager('two', function_test1, 'two', 3, 4)
        ThreadManager('three', function_test2, 'three', 5)
        ThreadManager.start_all_threads()
        print('main starting work at {}'.format(datetime.datetime.now().strftime(_time_format)))
        time.sleep(1)
        print('main done working at {}'.format(datetime.datetime.now().strftime(_time_format)))
        ThreadManager.join_all_threads()

        for thread in ThreadManager.threads:
            print(ThreadManager.threads[thread].rtn)
        print("Back to Main Thread")

        # Or you can manage them yourself:
        thread1 = ThreadManager('one', function_test1, 'one', 1, 2)
        thread2 = ThreadManager('two', function_test1, 'two', 3, 4)
        thread3 = ThreadManager('three', function_test2, 'three', 5)
        thread1.start()
        thread2.start()
        thread3.start()
        print('main starting work at {}'.format(datetime.datetime.now().strftime(_time_format)))
        time.sleep(1)
        print('main done working at {}'.format(datetime.datetime.now().strftime(_time_format)))
        thread1.join()
        thread2.join()
        thread3.join()
        print(thread1.rtn)
        print(thread2.rtn)
        print(thread3.rtn)
        print("Back to Main Thread")
        print('{} {} ending'.format(datetime.datetime.now(), current_function_name()))

    def test_worker_loop_thread(self):
        print('{} {} starting'.format(datetime.datetime.now(), current_function_name()))
        looper = WorkerLoopThread(worker_function)
        looper.start()
        time.sleep(5)
        looper.stop()
        print('{} {} ending'.format(datetime.datetime.now(), current_function_name()))



from thompcoutils.config_utils import ConfigManager
from thompcoutils.config_utils import EmailConnectionConfig
from thompcoutils.log_utils import get_logger
import smtplib
import os
from email import encoders
import mimetypes
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

if os.name == 'nt' or os.name == "posix":
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
else:
    # noinspection PyUnresolvedReferences
    from email.MIMEMultipart import MIMEMultipart
    # noinspection PyUnresolvedReferences
    from email.MIMEText import MIMEText
    # noinspection PyUnresolvedReferences
    from email.MIMEApplication import MIMEApplication


class EmailSender:
    def __init__(self, email_cfg):
        self.email_connection = email_cfg

    def send(self, to_email, subject, message, attach_files=None):
        logger = get_logger()
        logger.debug("sending to:{} subject:{}, message:{}, files:{}".format(to_email, subject, message, attach_files))
        if isinstance(attach_files, str):
            attach_files = [attach_files]
        logger = get_logger()
        server = None
        try:
            server = smtplib.SMTP(host=self.email_connection.smtp_host, port=self.email_connection.smtp_port)
            server.ehlo()
            server.starttls()
            server.login(self.email_connection.username, self.email_connection.password)
            outer_msg = MIMEMultipart('alternative')
            outer_msg['Subject'] = subject
            outer_msg.attach(MIMEText(message, "html"))
            outer_msg['From'] = self.email_connection.from_user
            if isinstance(to_email, list):
                outer_msg['To'] = ", ".join(to_email)
            else:
                outer_msg['To'] = to_email
            if attach_files is not None:
                for filename in attach_files:
                    file_type, encoding = mimetypes.guess_type(filename)
                    if file_type is None or encoding is not None:
                        file_type = 'application/octet-stream'
                    maintype, subtype = file_type.split('/', 1)
                    if maintype == 'text':
                        with open(filename) as fp:
                            msg = MIMEText(fp.read(), _subtype=subtype)
                    elif maintype == 'image':
                        with open(filename, 'rb') as fp:
                            msg = MIMEImage(fp.read(), _subtype=subtype)
                    elif maintype == 'audio':
                        with open(filename, 'rb') as fp:
                            msg = MIMEAudio(fp.read(), _subtype=subtype)
                    else:
                        with open(filename, 'rb') as fp:
                            msg = MIMEBase(maintype, subtype)
                            msg.set_payload(fp.read())
                        encoders.encode_base64(msg)
                    msg.add_header('Content-Disposition', 'attachment', filename=filename)
                    outer_msg.attach(msg)
            server.sendmail(self.email_connection.from_user, to_email, outer_msg.as_string())
            logger.debug("Successfully sent mail to " + str(to_email))
        except Exception as e:
            logger.warning("Failed to send mail to {} because {}".format(to_email, str(e)))
        finally:
            if server is not None:
                server.quit()

    def send_embedded_images(self, to_email, subject, html, files):
        msg_root = MIMEMultipart('related')
        msg_root['Subject'] = subject
        msg_root['From'] = self.email_connection.from_user
        msg_root['To'] = to_email
        msg_root.preamble = 'This is a multi-part message in MIME format.'

        msg_alternative = MIMEMultipart('alternative')
        msg_root.attach(msg_alternative)

        msg_text = MIMEText('Image(s)')
        msg_alternative.attach(msg_text)

        msg_text = MIMEText(html, 'html')
        msg_alternative.attach(msg_text)

        count = 1
        for file in files:
            fp = open(file, 'rb')
            msg_image = MIMEImage(fp.read())
            fp.close()
            msg_image.add_header('Content-ID', '<image{}>'.format(count))
            count += 1
            msg_root.attach(msg_image)

        import smtplib
        smtp = smtplib.SMTP(self.email_connection.smtp_host, self.email_connection.smtp_port)
        smtp.ehlo()
        smtp.starttls()
        smtp.login(self.email_connection.username, self.email_connection.password)
        smtp.sendmail(self.email_connection.from_user, to_email, msg_root.as_string())
        smtp.quit()


def main():
    logger = get_logger()
    create_file = True
    temp_filename = "testing EmailCfg only.ini"
    if create_file:
        if os.path.isfile(temp_filename):
            os.remove(temp_filename)
    cfg_mgr = ConfigManager(temp_filename, create=create_file)
    if create_file:
        EmailConnectionConfig(cfg_mgr=cfg_mgr)
        logger.debug("Writing out configuration file to {} and stopping.".format(temp_filename))
        cfg_mgr.write(temp_filename)
        raise Exception("Should never get here!")
    email_cfg = EmailConnectionConfig(cfg_mgr)
    mail_from_cfg = EmailSender(email_cfg)
    html_message = '<b>Some <i>HTML</i> text</b> and an image.<br>'\
                   '<img src="cid:image1">'\
                   '<img src="cid:image2">'\
                   '<br>Nifty!'

    mail_from_cfg.send_embedded_images("Jordan@ThompCo.com", "Testing", html_message,
                                       ["../attachments/image_1.png", "../attachments/image_2.png"])

    mail_from_cfg.send_embedded_images(to_email="Jordan@ThompCo.com", subject="this is a test from emailSender",
                                       html=html_message,
                                       files=["../attachments/image_1.png"])

    mail_from_cfg.send(to_email="Jordan@ThompCo.com", subject="this is a test from emailSender",
                       message="Here is the message using parameters passed in with an attached file",
                       attach_files=[os.path.basename(__file__)])
    mail_from_cfg.send(to_email="Jordan@ThompCo.com", subject="this is a test from emailSender",
                       message="Here is the message using parameters passed in")
    mail_from_cfg.send(to_email="Jordan@ThompCo.com", subject="this is a test from emailSender",
                       message="Here is the message using values from a configuration file with an attached file",
                       attach_files=[os.path.basename(__file__)])
    mail_from_cfg.send(to_email="Jordan@ThompCo.com", subject="this is a test from emailSender",
                       message="Here is the message using values from a configuration file")


if __name__ == "__main__":
    main()



import os
import logging
import unittest
import datetime
from thompcoutils.config_utils import ConfigManager
from thompcoutils.test_utils import assert_test
from thompcoutils import time_utils

test_path = 'test_ini_files'
if not os.path.exists(test_path):
    os.mkdir(test_path)
log_configuration_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.ini')
logging.config.fileConfig(log_configuration_file)


class TestTimeUtils(unittest.TestCase):
    def test_holidays(self):
        christmas = time_utils.Holiday(time_utils.Holiday.CHRISTMAS)
        assert_test(christmas.month == 12, "Should be December")
        assert_test(christmas.day == 25, "Should be 25")

        new_years = time_utils.Holiday(time_utils.Holiday.NEW_YEARS)
        assert_test(new_years.day == 1, "should be January")
        assert_test(new_years.month == 1, "should be 1")

        new_years = time_utils.Holiday("new years")
        assert_test(new_years.day == 1, "should be January")
        assert_test(new_years.month == 1, "should be 1")

        new_years = time_utils.Holiday("nEw yEaRs")
        assert_test(new_years.day == 1, "should be January")
        assert_test(new_years.month == 1, "should be 1")

        independence = time_utils.Holiday(time_utils.Holiday.INDEPENDENCE)
        assert_test(independence.month == 7, "should be July")
        assert_test(independence.day == 4, "should be 4")

        thanksgiving = time_utils.Holiday(time_utils.Holiday.THANKSGIVING)
        assert_test(thanksgiving.month == 11, "should be November")

        memorial = time_utils.Holiday(time_utils.Holiday.MEMORIAL)
        assert_test(memorial.month == 5, "should be November")
        assert_test(memorial.day == 31, "should be 31")

        labor= time_utils.Holiday(time_utils.Holiday.LABOR)
        assert_test(labor.month == 9, "should be September")

    def test_is_weekend(self):
        count = 0
        dt = datetime.datetime(year=datetime.datetime.now().year, month=1, day=1)
        for i in range(7):
            if time_utils.is_weekend(dt):
                count += 1
            dt = datetime.datetime(year=dt.year, month=1, day=dt.day+1)
        assert_test(count == 2, "can only be two weekend days in 7 days")




from thompcoutils.string_utils import list_to_string


class TestException(Exception):
    pass


def assert_test(value, message=None, error=None, passed=None):
    if message is not None:
        print(message, end="...")
    if value:
        if passed is None:
            print("passed")
        else:
            print(passed)
    else:
        if error is None:
            print("FAILED!!")
        else:
            print(error)
        assert False


def exception_test(message,  exception, method_to_test, **kwargs):
    print(message, end="...")
    try:
        method_to_test(**kwargs)
        print()
        raise TestException("{} test should have failed".format(message))
    except exception as e:
        if isinstance(e, TestException):
            raise e
        pass
    print("passed")


def _test(throws, ex):
    if throws:
        raise ex("Throws")


class FailedTestException(Exception):
    pass


def test(method, params, expected_value=None, expected_exception=None):
    method_name = str(method).split(" ")[1]
    print("starting {}({})...".format(method_name, list_to_string(params)), end="")
    if expected_exception is None:
        if expected_value is None:
            method(*params)
        elif method(*params) == expected_value:
            print("passed")
        else:
            str1 = "FAILED EXPECTED VALUE {}({})".format(method_name, list_to_string(params))
            raise FailedTestException(str1)
    else:
        try:
            if method(*params) == expected_value:
                str1 = "FAILED EXCEPTION {}({})".format(method_name, list_to_string(params))
                raise FailedTestException(str1)
        except expected_exception:
            print("passed")


def main():
    exception_test("testing - should throw 0", Exception, _test, throws=True, ex=Exception)
    try:
        exception_test("testing - should throw because the function didn't throw", TestException,
                       _test, throws=False, ex=TestException)
        raise Exception("this should not pass!")
    except TestException:
        pass
    try:
        exception_test("testing - should throw because exceptions don't match", Exception,
                       _test, throws=False, ex=TestException)
        raise Exception("this should not pass!")
    except TestException:
        pass
    assert_test(True, "just some testing 0")
    assert_test(True, "just some testing 1", "error occurred 1")
    assert_test(True, "just some testing 2", "error occurred 2", "this passed 2")
    # assert_test(False, "error occurred 3")
    # assert_test(False, "just some testing 4", "error occurred 4")
    # assert_test(False, "just some testing 5", "error occurred 5", "this passed 5")


if __name__ == '__main__':
    main()



import unittest
from thompcoutils import pio as pio
import time
import datetime

BUTTON_PIN = 16
TEMP_HUMIDITY_PIN = 4
LED_RED_PIN = 23
LED_GREEN_PIN = 24
LED_BLUE_PIN = 25

#######################################
# Run like this in the parent folder:
# For all tests:
#   python -m unittest tests/test_pio.py 
# For only one test:
#   python -m unittest tests.test_pio.TestPio.test_ColorLed
#######################################

def motor_callback(is_running, event_time):
    if is_running:
        print("motor started at:{}".format(event_time))
    else:
        print("motor stopped at:{}".format(event_time))


def button_callback(pin, state, argument):
    print('{} Pin {} now at {}.  Message:{}'.format(datetime.datetime.now(), pin, state, argument))


def timed_button_callback(pin, duration, argument):
    print('{} Pin {} now down for {}.  Message;{}'.format(datetime.datetime.now(), pin, duration, argument))


def button_light_squence_callback():
    pass


class TestPio(unittest.TestCase):
    def dont_test_ColorLed(self):
        print('Starting test_ColorLed')
        color_led = pio.ColorLed(LED_RED_PIN, LED_GREEN_PIN, LED_BLUE_PIN)
        print('RED')
        color_led.red()
        time.sleep(.1)
        print('GREEN')
        color_led.green()
        time.sleep(.1)
        print('BLUE')
        color_led.blue()
        time.sleep(.1)
        print('CYAN')
        color_led.cyan()
        time.sleep(.1)
        print('MAGENTA')
        color_led.magenta()
        time.sleep(.1)
        print('WHITE')
        color_led.white()
        time.sleep(.1)
        print('YELLOW')
        color_led.yellow()
        time.sleep(.1)
        print('OFF')
        color_led.turn_off()

        color_led.red()
        color_led.start_flashing([100, 200])
        time.sleep(5)
        color_led.stop_flashing()
        color_led.turn_off()

    def dont_test_led(self):
        print('Starting test_Led')
        led = pio.Led(5, False)
        while True:
            print('Turning on')
            led.turn_on()
            time.sleep(1)
            print('Turning off')
            led.turn_off()
            time.sleep(1)

    def test_Button(self):
        import RPi.GPIO as GPIO
        print('Starting test_Button')
        GPIO.remove_event_detect(BUTTON_PIN)
        pio.Button(pin=BUTTON_PIN, callback=button_callback)
        while True:
            time.sleep(1)

    def dont_test_TemperatureHumidity(self):
        temp_humid = pio.TemperatureHumidity(TEMP_HUMIDITY_PIN, pio.TemperatureHumidity.SensorType.AM2302)
        temp_humid.start()
        while True:
            if temp_humid.queue.empty():
                print('waiting for sensor to stabilize')
            while not temp_humid.queue.empty():
                values = temp_humid.queue.get()
                print('{time}\tTemp:{temp:6.2f}°\tHumidity:{humid:6.1f}%'.format(time=values.time_stamp,
                                                                                 temp=values.temperature,
                                                                                 humid=values.humidity))
            time.sleep(1)

    def dont_test_TimedButton(self):
        print('Starting test_TimedButton')
        pio.TimedButton(BUTTON_PIN, timed_button_callback)
        while True:
            time.sleep(1)

    def dont_test_accelerometer(self):
        acc = pio.Accelerometer()
        static_motor_values = pio.Motor.get_resting_values(acc, 2000)
        motor = pio.Motor(acc, static_motor_values)
        motor.start(motor_callback)

        while True:
            print("motor is running:{}".format(motor.is_moving()))
            time.sleep(.5)

    def dont_test_ADC(self):
        adc = pio.ADC(gain=1, r1=1000000, r2=220000, tolerance=.1, update=.1)
        adc.start()
        time.sleep(1)
        while True:
            while not adc.queue.empty():
                values = adc.queue.get()
                if values.values[0]:
                    print('{time_stamp}\t'
                          'a0:{v0:6.4f}V\t{a0:-6d}\t'
                          'a1:{v1:6.4f}V\t{a1:-6d}\t'
                          'a3:{v2:6.4f}V\t{a2:-6d}\t'
                          'a3:{v3:6.4f}V\t{a3:-6d}'.format(time_stamp=values.time_stamp,
                                                           v0=values.voltages[0], a0=values.values[0],
                                                           v1=values.voltages[1], a1=values.values[1],
                                                           v2=values.voltages[2], a2=values.values[2],
                                                           v3=values.voltages[3], a3=values.values[3]))
            time.sleep(.1)

    def dont_test_StateLight(self):
        color_led = pio.ColorLed(LED_RED_PIN, LED_GREEN_PIN, LED_BLUE_PIN)
        states = [pio.StateLight.State(light=color_led, colors=[False, True, True], on_off=[10, 10]),
                  pio.StateLight.State(light=color_led, colors=[True, False, True], on_off=[100, 10]),
                  pio.StateLight.State(light=color_led, colors=[True, True, False], on_off=[1000, 10]),
                  ]
        state_light = pio.StateLight(states)
        for loop in range(0, 10):
            state_light.state_stepper()
            time.sleep(5)
        color_led.stop_flashing()
        color_led.turn_off()

    def dont_test_button_light_sequence(self):
        color_led = pio.ColorLed(LED_RED_PIN, LED_GREEN_PIN, LED_BLUE_PIN)

        states = [pio.StateLight.State(light=color_led, colors=[False, True, True], on_off=[10, 10]),
                  pio.StateLight.State(light=color_led, colors=[True, False, True], on_off=[100, 10]),
                  pio.StateLight.State(light=color_led, colors=[True, True, False], on_off=[1000, 10]),
                  ]
        button_light_sequence = pio.ButtonLightSequence(callback=button_light_squence_callback, button_pin=BUTTON_PIN,
                                                        states=states, time_between_states=5000)
        time.sleep(900)



import threading
import time
import datetime
from thompcoutils.log_utils import get_logger

_time_format = '%H:%M:%S'


class ThreadManager (threading.Thread):
    THREAD_LOCK = threading.Lock()
    threads = {}

    def __init__(self, name, function, *argv):
        threading.Thread.__init__(self)
        self.name = name
        self.function = function
        self.args = argv
        self.rtn = None
        ThreadManager.threads[name] = self

    def run(self):
        self.rtn = self.function(*self.args)

    @staticmethod
    def start_thread(thread_name):
        ThreadManager.threads[thread_name].start_flashing()

    @staticmethod
    def start_all_threads():
        for thread_name in ThreadManager.threads:
            ThreadManager.threads[thread_name].start()

    @staticmethod
    def join_all_threads():
        for thread_name in ThreadManager.threads:
            ThreadManager.threads[thread_name].join()

    @staticmethod
    def join_thread(thread_name):
        ThreadManager.threads[thread_name].join()


class WorkerLoopThread(threading.Thread):
    def __init__(self, callback_function, parameters=None):
        super(WorkerLoopThread, self).__init__()
        logger = get_logger()
        logger.debug("creating a WorkerLoopThread")
        self.function = callback_function
        self.parameters = parameters
        self._is_running = False

    def is_running(self):
        return self._is_running

    def stop(self):
        self._is_running = False

    def start(self):
        if not self._is_running:
            self._is_running = True
            threading.Thread.start(self)

    def run(self):
        logger = get_logger()
        logger.debug("starting WorkerThread")
        while self._is_running:
            if self.parameters is None:
                logger.debug("calling function {}()".format(self.function))
                self.function()
            else:
                logger.debug("calling function {}({})".format(self.function, self.parameters))
                self.function(self.parameters)


class WorkerThread(threading.Thread):
    def __init__(self, callback_function, parameters=None):
        super(WorkerThread, self).__init__()
        logger = get_logger()
        logger.debug("creating a WorkerThread")
        self.function = callback_function
        self.parameters = parameters

    def run(self):
        logger = get_logger()
        logger.debug("starting WorkerThread")
        if self.parameters is None:
            logger.debug("calling function {}()".format(self.function))
            self.function()
        else:
            logger.debug("calling function {}({})".format(self.function, self.parameters))
            self.function(self.parameters)


class Watchdog(threading.Thread):
    def __init__(self, check_period, timeout_function, parameters):
        super(Watchdog, self).__init__()
        self.last_tickle = None
        self.is_running = False
        self.check_period = check_period
        self.timeout_function = timeout_function
        self.timeout_function_parameters = parameters

    def tickle(self):
        self.last_tickle = datetime.datetime.now()

    def stop(self):
        self.is_running = False

    def run(self):
        self.is_running = True
        self.tickle()
        while self.is_running:
            time.sleep(self.check_period)
            now = datetime.datetime.now()
            time_diff = now - self.last_tickle
            if time_diff.seconds > self.check_period:
                self.timeout_function(self.timeout_function_parameters)



from thompcoutils.log_utils import get_logger
import urllib.request
import json
import ssl
from werkzeug.exceptions import BadRequestKeyError


class MissingArgumentException(Exception):
    pass


class Server:
    @staticmethod
    def get_info(args, request):
        if args is None:
            args = []
        values = {}
        for arg in args:
            found = False
            if request.json is not None:
                try:
                    continue
                except BadRequestKeyError:
                    pass
            if not found:
                try:
                    values[arg] = str(request.args[arg])
                    continue
                except BadRequestKeyError:
                    pass
            if not found:
                try:
                    values[arg] = request.form.get(arg)
                    if values[arg] is not None:
                        continue
                except BadRequestKeyError:
                    pass
            if not found:
                try:
                    values[arg] = request.form[arg]
                    if values[arg] is not None:
                        continue
                except BadRequestKeyError:
                    pass
            if arg not in values:
                raise MissingArgumentException("argument '{}' not not found in {}".format(arg, request.url))
        return values


class Client:
    def __init__(self, host, port, page_root="", crt_file=None):
        logger = get_logger()
        self.host = host
        self.application_page = page_root
        self.port = port
        self.crt_file = crt_file
        logger.debug("host:{}, port:{}".format(host, port))
        if self.crt_file:
            http = "https"
        else:
            http = "http"
        if page_root is None:
            page_root = ""
        self.url = '{}://{}:{}/{}'.format(http, host, port, page_root)

    def send_curl(self, command, values):
        logger = get_logger()
        url = "{}/{}".format(self.url, command)
        logger.debug("Accessing URL: {}".format(url))
        req = urllib.request.Request(url)
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        json_data = json.dumps(values)
        json_data_bytes = json_data.encode('utf-8')  # needs to be bytes
        req.add_header('Content-Length', len(json_data_bytes))

        if self.crt_file:
            context = ssl.create_default_context()
            context.check_hostname = False
            context.load_verify_locations(self.crt_file)
            rtn = urllib.request.urlopen(req, json_data_bytes, context=context)
        else:
            rtn = urllib.request.urlopen(req, json_data_bytes)
        data = json.load(rtn)
        return data, rtn



import setuptools
import thompcoutils.version

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="thompcoutils",
    version=thompcoutils.version.version,
    author="Jordan Thompson",
    author_email="Jordan@ThompCo.com",
    description="Another collection of utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    install_requires=[
        'psutil',
        'netifaces',
        'python-dateutil'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)

exclude_package_data = {'': ['install.sh']},



from thompcoutils.log_utils import get_logger
from thompcoutils.cron_time import CronTime
import thompcoutils.file_utils as file_utils
from shutil import copyfile
import datetime
import sys
from builtins import staticmethod
import ast
from configparser import ConfigParser, NoOptionError, NoSectionError, DuplicateSectionError
import os
from dateutil import parser


class ConfigException(Exception):
    pass


class EmailConnectionConfig:
    def __init__(self, cfg_mgr, username='myname@google.com', password='mySecretPassword',
                 from_user='Where the email will come from', smtp_host='smtp.gmail.com',
                 smtp_port=587, section_heading='email connection',
                 username_tag='username', password_tag='password', from_tag='from', smtp_host_tag='smtp_host',
                 smtp_port_tag='smtp_port'):
        """
        This class represents data in a configuration file that represents an email configuration
        :param cfg_mgr: the configuration manager for this file
        :param username: email username
        :param password: email password
        :param from_user: user sending the email (could be an application)
        :param smtp_host: smtp host of the mail server
        :param smtp_port: smtp port of the mail server
        :param section_heading: heading of the email (for deprecated configuration files)
        :param username_tag: specialized username tag (for deprecated configuration files)
        :param password_tag: specialized password tag (for deprecated configuration files)
        :param from_tag: specialized from tag (for deprecated configuration files)
        :param smtp_host_tag: specialized smtp host tag (for deprecated configuration files)
        :param smtp_port_tag: specialzed smtp port tag (for deprecated configuration files)
        """
        self.username = cfg_mgr.read_entry(section=section_heading, entry=username_tag, default_value=username)
        self.password = cfg_mgr.read_entry(section=section_heading, entry=password_tag, default_value=password)
        self.from_user = cfg_mgr.read_entry(section=section_heading, entry=from_tag, default_value=from_user,
                                            notes='the user/application sending the email')
        self.smtp_host = cfg_mgr.read_entry(section=section_heading, entry=smtp_host_tag, default_value=smtp_host)
        self.smtp_port = cfg_mgr.read_entry(section=section_heading, entry=smtp_port_tag, default_value=smtp_port)


class HiLow:
    """
    HiLow class provides a config file to store low and hi values.  If a low value for a tag/entry is lower than the
    existing value, it will be updated.  Similarly, if a hi value for is higher than the existing hi value, it will be
    updated.
    """
    smallest = -sys.maxsize - 2
    biggest = sys.maxsize - 2
    hi_tag = 'hi'
    low_tag = 'low'
    last_tag = 'last'
    low_changed_tag = 'low_changed'
    hi_changed_tag = 'hi_changed'
    last_changed_tag = 'last_changed'
    direction_tag = 'direction'
    change_amount_tag = 'change_amount'

    class Direction(enumerate):
        Up = 'up'
        Down = 'down'
        NoChange = 'no change'

        @staticmethod
        def validate(string):
            if string == str(HiLow.Direction.Up):
                return HiLow.Direction.Up
            elif string == str(HiLow.Direction.Down):
                return HiLow.Direction.Down
            elif string == str(HiLow.Direction.NoChange):
                return HiLow.Direction.NoChange
            else:
                raise ConfigException('{} not recognized as a HiLow.Direction'.format(string))

    def __init__(self, file_name, value_type=float):
        """
        :param file_name: name of the file to store values
        """
        self.file_name = file_name
        if not os.path.exists(file_name):
            file_utils.touch(file_name)
        self.cfg_mgr = ConfigManager(file_name)
        self.value_type = value_type

    def read_values(self, entry):
        """
        gets the current values for a tag/entry
        :param entry: tag/entry`
        :return: a dictionary of values (i.e. {'hi': 10, 'low': 2} )
        """
        try:
            hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, value_type=self.value_type,
                                         default_value=self.smallest)
        except ValueError:
            hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, value_type=self.value_type,
                                         default_value=self.smallest + .01)
        try:
            low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, value_type=self.value_type,
                                          default_value=self.biggest)
        except ValueError:
            low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, value_type=self.value_type,
                                          default_value=self.biggest - .01)
        hi_changed_time = self.cfg_mgr.read_entry(section=entry, entry=self.hi_changed_tag,
                                                  default_value=datetime.datetime.now())
        low_changed_time = self.cfg_mgr.read_entry(section=entry, entry=self.low_changed_tag,
                                                   default_value=datetime.datetime.now())

        direction = self.cfg_mgr.read_entry(section=entry, entry=self.direction_tag,
                                            default_value=str(self.Direction.NoChange))
        direction = self.Direction.validate(direction)

        try:
            last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, value_type=self.value_type,
                                           default_value=0)
        except ValueError:
            last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, value_type=self.value_type,
                                           default_value=0.01)
        try:
            changed_amount = self.cfg_mgr.read_entry(section=entry, entry=self.change_amount_tag,
                                                     value_type=self.value_type, default_value=0)
        except ValueError:
            changed_amount = self.cfg_mgr.read_entry(section=entry, entry=self.change_amount_tag,
                                                     value_type=self.value_type, default_value=0.01)
        last_changed = self.cfg_mgr.read_entry(section=entry, entry=self.last_changed_tag,
                                               default_value=datetime.datetime.now())
        return {self.hi_tag: hi, self.low_tag: low,
                self.low_changed_tag: low_changed_time, self.hi_changed_tag: hi_changed_time,
                self.last_tag: last, self.last_changed_tag: last_changed, self.direction_tag: direction,
                self.change_amount_tag: changed_amount}

    def write_value(self, entry, value):
        try:
            self.cfg_mgr.config.add_section(entry)
        except DuplicateSectionError:
            pass

        try:
            cfg_last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, value_type=self.value_type,
                                               default_value=0)
        except ValueError:
            cfg_last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, value_type=self.value_type,
                                               default_value=.01)
        if value < cfg_last:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=cfg_last - value)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.Down)
            try:
                cfg_low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, value_type=self.value_type,
                                                  default_value=self.biggest)
            except ValueError:
                cfg_low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, value_type=self.value_type,
                                                  default_value=self.biggest - .01)
            if value < cfg_low:
                self.cfg_mgr.write_entry(section=entry, entry=self.low_changed_tag, value=datetime.datetime.now())
                self.cfg_mgr.write_entry(section=entry, entry=self.low_tag, value=value)
        elif value > cfg_last:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=value - cfg_last)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.Up)
            try:
                cfg_hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, value_type=self.value_type,
                                                 default_value=self.smallest)
            except ValueError:
                cfg_hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, value_type=self.value_type,
                                                 default_value=self.smallest + .01)
            if value > cfg_hi:
                self.cfg_mgr.write_entry(section=entry, entry=self.hi_changed_tag, value=datetime.datetime.now())
                self.cfg_mgr.write_entry(section=entry, entry=self.hi_tag, value=value)
        else:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=0)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.NoChange)
        self.cfg_mgr.write_entry(section=entry, entry=self.last_tag, value=value)
        self.cfg_mgr.write_entry(section=entry, entry=self.last_changed_tag, value=datetime.datetime.now())
        self.cfg_mgr.write(out_file=self.file_name, stop=False, overwrite=True)
        return self.read_values(entry)


class ConfigManager:
    def __init__(self, file_name, title=None, create=False):
        """
        ConfigManager allows for reading and writing a configuration file
        :param file_name: name of the config file
        :param title: information to put at the top of the file when writing it out
        :param create: if True, creates the file
        """
        self.file_name = file_name
        self.config = ConfigParser()
        self.config.optionxform = str
        self.create = create
        if not create:
            if os.path.exists(file_name):
                self.config.read(file_name)
            else:
                raise FileNotFoundError('File {} does not exist!'.format(file_name))
        self.notes = []
        self.title = title
        self.values = {}

    @staticmethod
    def missing_entry(section, entry, file_name, default_value=None):
        """
        this method logs a message about an entry that is missing from the config file
        :param section: section of the config file
        :param entry: entry in the section of the config file
        :param file_name: name of the file
        :param default_value: default value for the entry
        :return: None
        """
        logger = get_logger()
        logger.debug('starting')
        if default_value is None:
            log_fn = logger.critical
            message = 'Required entry'
            default_value = ''
        else:
            log_fn = logger.debug
            message = 'Entry'
            if default_value == '':
                default_value = 'Ignoring.'
            else:
                default_value = 'Using default value of (' + str(default_value) + ')'
        log_fn(message + ' \"' + entry + '\" in section [' + section + '] in file: ' + file_name
               + ' is malformed or missing.  ' + str(default_value))
        if default_value == '':
            log_fn('Exiting now')
            sys.exit()

    @staticmethod
    def _insert_note(lines, line_number, note):
        """
        This method inserts a note at a particular line in an array of lines that will ultimately be written to the
        config file
        :param lines: array of strings
        :param line_number: the line number to insert the note at
        :param note: the note to be inserted
        :return:
        """
        if '\n' in note:
            message = note.split('\n')
        else:
            message = note
        if message is None:
            pass
        elif type(message) == str:
            lines.insert(line_number, '# ' + message + ':\n')
        else:
            for line in message[:-1]:
                lines.insert(line_number, '# ' + line + '\n')
                line_number += 1
            lines.insert(line_number, '# ' + message[-1] + ':\n')
        lines[line_number + 1] += "\n"

    def add_note(self, section, entry, notes):
        """
        Adds a note to an entry or a section (if entry is None)
        :param section: Section to add the note to
        :param notes: note to add (long lines will be broken up)
        :param entry: entry to add the note to.  If this is None, the note is added to the section itself.
        :return: None
        """
        self.notes.append({'section': section,
                           'entry': entry,
                           'notes': notes})

    def read_entry(self, section, entry, default_value, notes=None, value_type=None,
                   use_default_if_missing=True, allowed_values=None):
        """
        This method reads an entry in the config file
        :param section: section of the config file
        :param entry: entry in the section of the config file
        :param default_value: default value to use if the entry is missing
        :param notes: notes about the entry
        :param value_type: the type of data the entry represents
        :param use_default_if_missing: if True, the default value is used if the entry is missing
        :param allowed_values: only allow one of the entries in this list
        :return: the value in the entry
        """
        logger = get_logger()
        value = default_value
        wrong_type = None
        if self.create:
            try:
                self.config.add_section(section)
            except DuplicateSectionError:
                pass
            if notes:
                if allowed_values:
                    notes += " (must be one of " + str(allowed_values) + ")"
                self.add_note(section, entry, notes)
            self.config.set(section, entry, str(default_value))
        else:
            if value_type is None:
                if default_value is None:
                    raise ConfigException('if default_value=None, value_type must be set')
                value_type = type(default_value)
            try:
                if value_type == str:
                    value = self.config.get(section, entry)
                    if not isinstance(value, str):
                        wrong_type = str
                elif value_type == bool:
                    value = self.config.getboolean(section, entry)
                    if not isinstance(value, bool):
                        wrong_type = bool
                elif value_type == int:
                    value = self.config.getint(section, entry)
                    if not isinstance(value, int):
                        wrong_type = int
                elif value_type == float:
                    value = self.config.getfloat(section, entry)
                    if not isinstance(value, float):
                        wrong_type = float
                elif value_type == dict:
                    value = ast.literal_eval(self.config.get(section, entry))
                    if not isinstance(value, dict):
                        wrong_type = dict
                elif value_type == list:
                    value = ast.literal_eval(self.config.get(section, entry))
                    if not isinstance(value, list):
                        wrong_type = list
                elif value_type == datetime.datetime:
                    value = parser.parse(self.config.get(section, entry))
                    if not isinstance(value, datetime.datetime):
                        wrong_type = datetime.datetime
                elif value_type == CronTime:
                    format_entry = '{}_format'.format(entry)
                    time_entry = '{}_time'.format(entry)
                    try:
                        time_format = self.config.get(section, format_entry)
                    except NoOptionError:
                        raise ConfigException('{} missing for entry {} under section {}', format_entry, entry, section)
                    try:
                        value = self.config.get(section, time_entry)
                    except NoOptionError:
                        raise ConfigException('{} missing for entry {} under section {}', time_entry, entry, section)
                    value = CronTime.strfpcrontime(value, time_format)
                    if not isinstance(value, CronTime):
                        wrong_type = CronTime
                elif value_type == datetime.timedelta:
                    value = self.config.get(section, entry)
                    from pytimeparse import timeparse
                    seconds = timeparse.timeparse(value)
                    value = datetime.timedelta(seconds=seconds)
                elif value_type == datetime.time:
                    value = self.config.get(section, entry)
                    value = datetime.datetime.strptime(value, '%H:%M:%S').time()
                else:
                    raise ConfigException('type {} not handled for ()'.format(type(default_value), default_value))
            except NoOptionError:
                logger.debug('Entry {} in section [{}] is missing.  Using default value of {}'.format(entry, section,
                                                                                                      default_value))
                if not use_default_if_missing:
                    value = None
            except NoSectionError:
                logger.debug('section [{}] is missing.  Using default value of {}'.format(section, default_value))
                if not use_default_if_missing:
                    value = None
            except ValueError as e:
                raise ConfigException("Invalid value of {} ({}) for entry {} under section {}".
                                      format(value, str(e), entry, section))

        if wrong_type is not None:
            raise ConfigException("{} is not a {} for entry {} under section {}".
                                  format(value, str(wrong_type), entry, section))

        if allowed_values and value not in allowed_values:
            raise ConfigException("{} is not one of the allowed values: {} for entry {} under section {}".
                                  format(value, allowed_values, entry, section))
        return value

    def read_section(self, section, default_entries, notes=None):
        """
        This method reads an entire section
        :param section:
        :param default_entries:
        :param notes:
        :return:
        """
        key_values = default_entries
        if self.create:
            try:
                self.config.add_section(section)
            except DuplicateSectionError:
                pass
            for entry in default_entries:
                self.config.set(section, str(entry), str(default_entries[entry]))
            if notes is not None:
                self.add_note(section, notes)
        else:
            key_values = dict()
            for (key, val) in self.config.items(section):
                key_values[key] = val
        return key_values

    def add_section_notes(self, section, notes):
        self.notes.append({'section': section,
                           'entry': None,
                           'notes': notes})

    def write_entry(self, section, entry, value, note=None, format_string=None):
        """
        This method writes the information into a member array that will be written to the file later
        :param section: section of the config file
        :param entry: entry in the section of the config file
        :param value: value for the entry
        :param note: any notes for this entry
        :param format_string: special formatting to converty the value to a string representation
        :return:
        """
        try:
            if isinstance(value, CronTime):
                self.config.set(section, '{}_format'.format(entry), format_string.replace('%', '%%'))
                self.config.set(section, '{}_time'.format(entry), value.strfcrontime(format_string))
            else:
                self.config.set(section, str(entry), str(value))
        except DuplicateSectionError:
            self.config.add_section(section)
            self.config.set(section, str(entry), str(value))

        if note is not None:
            self.add_note(section, entry, note)

    def write(self, *args, **kwargs):
        overwrite = kwargs.get('overwrite', False)
        out_file = kwargs.get('out_file', True)  # stop=True, overwrite=False):
        stop = kwargs.get('stop', True)  # stop=True, overwrite=False):
        if os.path.isfile(out_file) and not overwrite:
            raise ConfigException('File {} exists!  You must remove it before running this'.format(out_file))
        f = open(out_file, 'w')
        self.config.write(f)
        f.close()
        f = open(out_file)
        lines = f.readlines()
        f.close()
        if self.title is not None:
            ConfigManager._insert_note(lines, 0, self.title)
        for note in self.notes:
            in_section = False
            line_number = 0
            for line in lines:
                if line.startswith("[" + note["section"] + "]"):
                    in_section = True
                elif in_section and line.startswith(note['entry']):
                    ConfigManager._insert_note(lines, line_number, note['notes'])
                    break
                line_number += 1
        f = open(out_file, 'w')
        contents = ''.join(lines)
        f.write(contents)
        f.close()
        if stop:
            print('Done writing {} configuration file.  Stopping execution, please re-run'.format(out_file))
            sys.exit()


class PiConfig(ConfigManager):
    _boot_folder = '/boot'
    _wpa_supplicant_file = '/etc/wpa_supplicant/wpa_supplicant.conf'
    _host_name_file = '/etc/hostname'
    _hosts_file = '/etc/hosts'
    _default_file_extension = '.default'

    class WiFi:
        section = 'wifi'
        ssid_tag = 'ssid'
        key_tag = 'key'
        hidden_tag = 'hidden'
        priority_tag = 'priority'

        def __init__(self, cfg_mgr, ssid='my wifi ap', wifi_key='my wifi key', hidden=False, section=section,
                     priority=1):
            """
            WifiConfig defines the data for the wpa_suplicant file
            :param cfg_mgr: the ConfigManager
            :param ssid: SSID for this wifi
            :param wifi_key: password key for this wifi
            :param hidden: indicates if this wifi is hidden
            :param section: normally this would be 'wifi', but if there are multiple wifi sections, they must start
            with 'wifi'
            """
            self.section = section
            cfg_mgr.add_note(section, 'More than one ssid can be specified by adding sections that begin '
                                      'with "{}"'.format(PiConfig.WiFi.section))

            self.ssid = cfg_mgr.read_entry(section, PiConfig.WiFi.ssid_tag, ssid)
            self.wifi_key = cfg_mgr.read_entry(section, PiConfig.WiFi.key_tag, wifi_key)
            self.hidden = cfg_mgr.read_entry(section, PiConfig.WiFi.hidden_tag, hidden)
            self.priority = cfg_mgr.read_entry(section, PiConfig.WiFi.priority_tag, priority,
                                               notes='useful if more than one ssid is specified')

    class Miscellaneous:
        section = 'misc'
        host_name_tag = 'host_name'
        location_tag = 'location'
        wifi_country_code_tag = 'wifi_country_code'

        def __init__(self, cfg_mgr, host_name='Pi', wifi_country_code='US', location='Office'):
            """
            WifiConfig defines the data for the wpa_suplicant file
            :param cfg_mgr: the ConfigManager
            :param host_name: name for this host
            :param wifi_country_code: country code for the wifi networks
            :param location: location of this host.  This is a place-holder for other applications to access
            """
            self.host_name = cfg_mgr.read_entry(PiConfig.Miscellaneous.section,
                                                PiConfig.Miscellaneous.host_name_tag, host_name)
            self.wifi_country_code = cfg_mgr.read_entry(PiConfig.Miscellaneous.section,
                                                        PiConfig.Miscellaneous.wifi_country_code_tag,
                                                        wifi_country_code)
            self.location = cfg_mgr.read_entry(PiConfig.Miscellaneous.section,
                                               PiConfig.Miscellaneous.location_tag,
                                               location)

    def __init__(self,
                 file_name='configuration.ini',
                 title=None,
                 default_filename='configuration.ini{}'.format(_default_file_extension)):
        """
        PiConfig extends ThompCoUtils.config_utils.ConfigManager.  It reads the specified file in the root directory
        (when viewed from another OS as an SD card or in the /boot folder when viewed from the pi).  It will create the
        specified default_filename in the boot_folder.  If file_name exists, it will read it and delete it.
        NOTE that this class can only be used on a raspberry pi!
        :param file_name: name of the file to read
        :param title:
        :param default_filename: filename to create if it is not already there
        """
        if not os.path.isdir(PiConfig._boot_folder):
            raise RuntimeError('This can only be used on a Raspberry Pi')
        self.default_filename = os.path.join(PiConfig._boot_folder, default_filename)
        file_name = os.path.join(PiConfig._boot_folder, file_name)
        self.wifi_sections = []
        self.misc = None
        self.read_config = False
        # try to open the actual file first
        if os.path.exists(file_name):
            super().__init__(file_name=file_name, title=title, create=False)
            for section in self.config.sections():
                if section.startswith(PiConfig.WiFi.section):
                    self.wifi_sections.append(PiConfig.WiFi(self, section=section))
            self.misc = PiConfig.Miscellaneous(self)
            # read the file, now delete it
            os.unlink(self.file_name)
            self.read_config = True
        if not os.path.exists(self.default_filename):
            # now write out the sample one
            super().__init__(file_name=self.default_filename, title=title, create=True)
            self.wifi_sections.append(PiConfig.WiFi(self, section='{}_first'.format(PiConfig.WiFi.section),
                                                    ssid='ssid 1', wifi_key='password 1',
                                                    priority=1))
            self.wifi_sections.append(PiConfig.WiFi(self, section='{}_second'.format(PiConfig.WiFi.section),
                                                    ssid='ssid 2', wifi_key='password 2',
                                                    priority=2))
            PiConfig.Miscellaneous(self)

    @staticmethod
    def backup_original_file(file_name):
        original_backup = "{}.original".format(file_name)
        if not os.path.exists(original_backup):
            copyfile(file_name, original_backup)

    def write_wpa_supplicant(self):
        if self.read_config:
            PiConfig.backup_original_file(PiConfig._wpa_supplicant_file)
            f = open(PiConfig._wpa_supplicant_file, 'w')
            f.write('Country={}\n'.format(self.misc.wifi_country_code))
            f.write('ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev\n')
            f.write('update_config=1\n')
            f.write('\n')
            for section in self.wifi_sections:
                f.write('network = {\n')
                f.write('\tssid = "{}"\n'.format(section.ssid))
                f.write('\tpsk = "{}"\n'.format(section.wifi_key))
                if section.hidden:
                    f.write('\tscan_ssid = 1\n')
                f.write('}\n')
                f.write('\n')
            f.close()

    def write_host_files(self):
        host_ip = '127.0.0.1'
        if self.read_config:
            PiConfig.backup_original_file(PiConfig._host_name_file)
            f = open(PiConfig._host_name_file, 'w')
            f.write('{}\n'.format(self.misc.host_name))
            f.close()
            PiConfig.backup_original_file(PiConfig._hosts_file)
            with open(PiConfig._hosts_file, 'r') as file:
                lines = file.readlines()
            for i in range(0, len(lines) - 1):
                if host_ip in lines[i]:
                    lines[i] = '{}\t{}'.format(host_ip, self.misc.host_name)
            with open(PiConfig._hosts_file, 'w') as file:
                file.writelines(lines)

    def write(self, *args, **kwargs):
        """
        This method writes out the default file.
        :param args:
        :param kwargs:
        :return:
        """
        super(PiConfig, self).write(out_file=self.default_filename, stop=False, overwrite=True)
